#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/queue.h>

#include "dtsadmin_nodeinfo_handler.h"

static
int str2ninfo(const char * str, struct node_info * ninfo)
{
	char * slash_tp = NULL, * slash_sp = NULL,
		* percent_tp = NULL, * sharp_tp = NULL,
		* sharp_sp = NULL, * percent_sp = NULL, * work = NULL;

	if (str == NULL || strlen(str) == 0 || ninfo == NULL) {
		return -1;
	}

	if ((work = strdup(str)) == NULL) {
		return -1;
	}

	/* get the hb_node_name and add. */
	slash_tp = strtok_r(work, TOKEN_SLASH, &slash_sp);
	if (slash_tp != NULL) {
		/* get the hb_node_name and ipaddr. */
		percent_tp = strtok_r(slash_tp, TOKEN_PERCENT, &percent_sp);
		if (percent_tp != NULL) {
			/* get the hb_node_name. */
			strncpy(ninfo->hb_node_name, percent_tp, MAX_LEN_HOST);
		}
		percent_tp = strtok_r(NULL, TOKEN_PERCENT, &percent_sp);
		if (percent_tp != NULL) {
			/* get the internal ip addr. */
			strncpy(ninfo->addr, percent_tp, MAX_LEN_HOST);
		}
	}

	/* get the vnode id and vnode id hash value. */
	slash_tp = strtok_r(NULL, TOKEN_SLASH, &slash_sp);
	if (slash_tp != NULL) {
		/* get the vnode id. */
		sharp_tp = strtok_r(slash_tp, TOKEN_SHARP, &sharp_sp);
		if (sharp_tp != NULL) {
			strncpy(ninfo->id, sharp_tp, MAX_LEN_VNODE_ID);
		}
		/* get the vnode id hash value. */
		sharp_tp = strtok_r(NULL, TOKEN_SHARP, &sharp_sp);
		if (sharp_tp != NULL) {
			strncpy(ninfo->id_hash, sharp_tp, MAX_LEN_VNODE_HASH_ID);
		}
	}

	free(work);
	return 0;
}

static
int add_ninfo(const char * str, struct node_info_head * ni_head)
{
	struct node_info * to_add = NULL;

	to_add = (struct node_info *) malloc(sizeof(struct node_info));
	if (to_add == NULL) {
		goto free4fail;
	} else {
		memset(to_add, 0, sizeof(struct node_info));
	}

	if (str2ninfo(str, to_add) != 0) {
		goto free4fail;
	}

	TAILQ_INSERT_TAIL(ni_head, to_add, entry);

	return 0;

free4fail:
	free(to_add);
	to_add = NULL;
	return -1;
}

/* expect following format.                           */
/* dts01/1#00000001:dts02/#00000003:dts03:dts04(dead) */
int str2ninfoq(const char * str, struct node_info_head * ni_head)
{
	int ret = 0;
	char * tp = NULL, * sp = NULL, * work = NULL;

	if (str == NULL || strlen(str) == 0 || ni_head == NULL) {
		return -1;
	}

	TAILQ_INIT(ni_head);

	if ((work = strdup(str)) == NULL) {
		return -1;
	}

	tp = strtok_r(work, TOKEN_COLON, &sp);
	if (tp != NULL) {
		if (add_ninfo(tp, ni_head) != 0) {
			/* nothing to do. */
		}
	}

	while (tp != NULL) {
		tp = strtok_r(NULL, TOKEN_COLON, &sp);
		if (tp != NULL) {
			if (add_ninfo(tp, ni_head) != 0) {
				/* nothing to do. */
				continue;
			}
		}
	}

	ret = 0;

	free(work);
	return 0;
}

static
int add_ninfo_vidset(struct node_info_vidset_head * nivhead,
			const struct node_info * ni)
{
	// found flag of node name.
	int is_found = 0;
	struct node_info_vidset * niv;

	for (niv = TAILQ_FIRST(nivhead); niv != NULL; niv = TAILQ_NEXT(niv, entry)) {
		if (strcmp(niv->hb_node_name, ni->hb_node_name) == 0) {
			// the node name was found. add the vnode id to current vid set.
			append_IDset(atoi(ni->id), niv->idset);
			is_found = 1;
		}
	}

	// new node name comming.
	if (is_found == 0) {
		struct node_info_vidset * to_add;
		to_add = (struct node_info_vidset *) malloc(sizeof(struct node_info_vidset));
		if (to_add == NULL) {
			goto free4fail;
		}
		memset(to_add, 0, sizeof(struct node_info_vidset));
		// copy the node name.
		if (ni->hb_node_name != NULL && strlen(ni->hb_node_name) != 0) {
			strncpy(to_add->hb_node_name, ni->hb_node_name, MAX_LEN_HOST);
		}
		// copy the ipaddr.
		if (ni->addr != NULL && strlen(ni->addr) != 0) {
			strncpy(to_add->addr, ni->addr, MAX_LEN_HOST);
		}
		append_IDset(atoi(ni->id), to_add->idset);
		TAILQ_INSERT_TAIL(nivhead, to_add, entry);
	}

	return 0;
free4fail:
	return -1;
}

int ninfoq2ninfo_vidsetq(struct node_info_head * ni_head,
		   struct node_info_vidset_head * niv_head)
{
	struct node_info * ni;

	TAILQ_INIT(niv_head);
	for (ni = TAILQ_FIRST(ni_head); ni != NULL; ni = TAILQ_NEXT(ni, entry)) {
		if (add_ninfo_vidset(niv_head, ni) != 0) {
			return -1;
		}
	}
	return 0;
}

int free_ninfo_vidsetq(struct node_info_vidset_head * niv_head)
{
	struct node_info_vidset * niv = NULL;

	while ((niv = TAILQ_FIRST(niv_head)) != NULL) {
		TAILQ_REMOVE(niv_head, niv, entry);
		free(niv);
		niv = NULL;
	}

	return 0;
}

int free_ninfoq(struct node_info_head * ni_head)
{
	struct node_info * ninfo = NULL;

	while ((ninfo = TAILQ_FIRST(ni_head)) != NULL) {
		TAILQ_REMOVE(ni_head, ninfo, entry);
		free(ninfo);
		ninfo = NULL;
	}

	return 0;
}


void print_ninfo(const struct node_info_head * ni_head)
{
	struct node_info * ninfo = NULL;

	for (ninfo = TAILQ_FIRST(ni_head); ninfo != NULL; ninfo = TAILQ_NEXT(ninfo, entry)) {
		printf("hb_node_name: %s\n", ninfo->hb_node_name);
		printf("id: %s\n", ninfo->id);
		printf("id hash: %s\n", ninfo->id_hash);
	}
}

/* the test driver */
/*
int main(int argc, char * argv[]) {
	struct node_info_head ni_head;

	if (argc != 2) {
		fprintf(stderr, "invalid arguments.\n");
		exit(1);
	}

	if (str2ninfoq(argv[1], &ni_head) != 0) {
		fprintf(stderr, "failed to str2ninfos\n");
	}
	print_ninfo(&ni_head);
	free_ninfoq(&ni_head);
	return 0;
}
*/

