#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/queue.h>

/* use for return values length. */
#include "common.h"
#include "dtsadmin_codes.h"
#include "dtsadmin_result_handler.h"

static
int trim_line_feed_code(char * str)
{
        int str_len = 0;

        if (str == NULL) {
                return -1;
        }
        str_len = strlen(str);
        if (str[str_len - 1] == '\n') {
                str[str_len - 1] = '\0';
        }
        return 0;
}

static
int str2result(const char * result_str, struct processed_results * to_add)
{
	char * host_tp = NULL;
	char * stat_tp = NULL;
	char * err_tp = NULL;
	char * msg_tp = NULL;

	if ((host_tp = strstr(result_str, TOKEN_HOST)) == NULL) {
		return -1;
	}
	if ((stat_tp = strstr(result_str, TOKEN_STAT)) == NULL) {
		return -1;
	}
	if ((err_tp = strstr(result_str, TOKEN_ERR)) == NULL) {
		return -1;
	}
	if ((msg_tp = strstr(result_str, TOKEN_MSG)) == NULL) {
		return -1;
	}

	strncpy(to_add->host, host_tp + strlen(TOKEN_HOST),
			stat_tp - (host_tp + strlen(TOKEN_HOST)));
	trim_line_feed_code(to_add->host);

	strncpy(to_add->stat, stat_tp + strlen(TOKEN_STAT),
			err_tp - (stat_tp + strlen(TOKEN_STAT)));
	trim_line_feed_code(to_add->stat);

	strncpy(to_add->error, err_tp + strlen(TOKEN_ERR),
			msg_tp - (err_tp + strlen(TOKEN_ERR)));
	trim_line_feed_code(to_add->error);

	strncpy(to_add->msg, msg_tp + strlen(TOKEN_MSG), MAX_LEN_RES);
	trim_line_feed_code(to_add->msg);

	return 0;
}

/* convert the string arrs to result queue. */
int strarr2resultq(const char * result_strv[], int result_cnt, struct results_head * rhead)
{
	int i = 0;
	struct processed_results * to_add = NULL;

	for (i = 0; i < result_cnt; i++) {
		if (result_strv[i] == NULL) {
			return -1;
		}

		to_add = (struct processed_results *) malloc(sizeof(struct processed_results));
		if (to_add == NULL) {
			return -1;
		} else {
			memset(to_add, 0, sizeof(struct processed_results));
		}

		if (str2result(result_strv[i], to_add) != 0) {
			free(to_add);
			return -1;
		}

		TAILQ_INSERT_TAIL(rhead, to_add, entry);
	}

	return 0;
}

int free_resultq(struct results_head * rhead)
{
	struct processed_results * result = NULL;

	while ((result  = TAILQ_FIRST(rhead)) != NULL) {
		TAILQ_REMOVE(rhead, result, entry);
		free(result);
		result = NULL;
	}

	return 0;
}

int judge_resultq(struct results_head * rhead)
{
	struct processed_results * result = NULL;

	for (result = TAILQ_FIRST(rhead); result != NULL; result = TAILQ_NEXT(result, entry)) {
		if (strncmp(TOKEN_OK, result->stat, strlen(TOKEN_OK)) != 0) {
			eprintf(HERE "host: %s, error: %s\n", result->host, result->error);
			return -1;
		}
	}

	return 0;
}

void print_resultq(struct results_head * rhead)
{
	struct processed_results * result = NULL;

	for (result = TAILQ_FIRST(rhead); result != NULL; result = TAILQ_NEXT(result, entry)) {
		printf("host: %s\n", result->host);
		printf("stat: %s\n", result->stat);
		printf("error: %s\n", result->error);
		printf("msg: %s\n", result->msg);
	}
}

