#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/queue.h>

/* use for return values length. */
#include "common.h"
#include "dtsadmin_server_conf_handler.h"

static
int trim_line_feed_code(char * str)
{
	int str_len = 0;

	if (str == NULL) {
		return -1;
	}
	str_len = strlen(str);
	if (str[str_len - 1] == '\n') {
		str[str_len - 1] = '\0';
	}
	return 0;
}

static
int set_sconf_val(const char * str, struct server_confs * sconf)
{
	int ret = 0;
	char * work = NULL;
	char * key = NULL;
	char * val = NULL;

	if ((work = strdup(str)) == NULL) {
		return -1;
	}

	key = strtok(work, TOKEN_EQUAL);
	if (key == NULL) {
		ret = -1;
		goto free4fail;
	}

	val = strtok(NULL, TOKEN_EQUAL);
	if (val == NULL) {
		/* only key specified.  */
		ret = 0;
		goto free4success;
	}
	if (strncmp(TOKEN_SRV_ADDR, key, strlen(TOKEN_SRV_ADDR)) == 0) {
		strncpy(sconf->srv_addr, val, MAX_LEN_SRV_ADDR);
		trim_line_feed_code(sconf->srv_addr);
	} else if (strncmp(TOKEN_SRV_PORT, key, strlen(TOKEN_SRV_PORT)) == 0) {
		strncpy(sconf->srv_port, val, MAX_LEN_SRV_PORT);
		trim_line_feed_code(sconf->srv_port);
	} else if (strncmp(TOKEN_IFACE, key, strlen(TOKEN_IFACE)) == 0) {
		strncpy(sconf->iface, val, MAX_LEN_IFACE);
		trim_line_feed_code(sconf->iface);
	} else if (strncmp(TOKEN_NETMASK, key, strlen(TOKEN_NETMASK)) == 0) {
		strncpy(sconf->netmask, val, MAX_LEN_NETMASK);
		trim_line_feed_code(sconf->netmask);
	} else if (strncmp(TOKEN_FAKEMAC, key, strlen(TOKEN_FAKEMAC)) == 0) {
		strncpy(sconf->fakemac, val, MAX_LEN_FAKEMAC);
		trim_line_feed_code(sconf->fakemac);
	} else if (strncmp(TOKEN_RI_OPT, key, strlen(TOKEN_RI_OPT)) == 0) {
		strncpy(sconf->ri_opt, val, MAX_LEN_RI_OPT);
		trim_line_feed_code(sconf->ri_opt);
	} else if (strncmp(TOKEN_RI, key, strlen(TOKEN_RI)) == 0) {
		strncpy(sconf->ri, val, MAX_LEN_RI);
		trim_line_feed_code(sconf->ri);
	} else if (strncmp(TOKEN_MACADDR, key, strlen(TOKEN_MACADDR)) == 0) {
		strncpy(sconf->macaddr, val, MAX_LEN_MACADDR);
		trim_line_feed_code(sconf->macaddr);
	} else if (strncmp(TOKEN_NODES, key, strlen(TOKEN_NODES)) == 0) {
		strncpy(sconf->nodes, val, MAX_LEN_RETURNED_NODES);
		trim_line_feed_code(sconf->nodes);
	} else if (strncmp(TOKEN_PROXY_PORT, key, strlen(TOKEN_PROXY_PORT)) == 0) {
		strncpy(sconf->proxy_port, val, MAX_LEN_PROXY_PORT);
		trim_line_feed_code(sconf->proxy_port);
	}

	ret = 0;
	goto free4success;

free4fail:
free4success:
	free(work);
	return ret;
}

static
int str2sconf(const char * msg, struct server_confs * to_add)
{
	int ret = 0;
	char * work = NULL;
	char * tp = NULL;
	char * sp = NULL;

	if (msg == NULL || strnlen(msg, MAX_MSG_SIZE) == 0) {
		return -1;
	}

	if ((work = strdup(msg)) == NULL) {
		return -1;
	}
	tp = strtok_r(work, TOKEN_QUESTION, &sp);
	if (tp != NULL) {
		if (set_sconf_val(tp, to_add) != 0) {
			ret = -1;
			goto free4fail;
		}
	}
	while (tp != NULL) {
		tp = strtok_r(NULL, TOKEN_QUESTION, &sp);
		if (tp != NULL) {
			if (set_sconf_val(tp, to_add) != 0) {
				ret = -1;
				goto free4fail;
			}
		}
	}

	ret = 0;
	goto free4success;

free4fail:
free4success:
	free(work);
	return ret;
}

static
int add_sconf(const char * conf_str, struct server_confs_head * shead)
{
	int ret = 0;
	struct server_confs * to_add = NULL;

	if (conf_str == NULL) {
		return -1;
	}

	to_add = (struct server_confs *) malloc(sizeof(struct server_confs));
	if (to_add == NULL) {
		return -1;
	} else {
		memset(to_add, 0, sizeof(struct server_confs));
	}

	if (str2sconf(conf_str, to_add) != 0) {
		ret = -1;
		goto free4fail;
	}

	TAILQ_INSERT_TAIL(shead, to_add, entry);

	ret = 0;
	goto free4success;
free4fail:
	free(to_add);
	to_add = NULL;
	return ret;
free4success:
	return ret;
}

/* parse message and value set to queue. */
int str2sconfq(const char * str, struct server_confs_head * shead)
{
	char srv_conf_str[MAX_LEN_RETURNED_SRV_CONFS] = "";
	char token[WORK_BUF_SIZE] = "";
	char * start_tp = NULL;
	char * end_tp = NULL;

	/* copy count to srv_conf_str. */
	int cp_cnt = 0;
	/* total copy count. */
	int total_cp_cnt = 0;

	if (str == NULL || strnlen(str, MAX_MSG_SIZE) == 0) {
		return -1;
	}

	strcat(token, TOKEN_QUESTION);
	strcat(token, TOKEN_SRV_ADDR);

	cp_cnt = 0;
	total_cp_cnt = 0;
	/* get the each server conf str and convert it to struct server_confs. */
	while (1) {
		/* search the first token. */
		if ((start_tp = strstr(str + total_cp_cnt, token)) == NULL) {
			/* nothing to add. */
			break;
		}
		/* search the next token. */
		if ((end_tp = strstr(start_tp + 1, token)) == NULL) {
			strncpy(srv_conf_str, start_tp, MAX_LEN_RETURNED_SRV_CONFS);
			add_sconf(srv_conf_str, shead);
			memset(srv_conf_str, 0, MAX_LEN_RETURNED_SRV_CONFS);
			// all srv conf was added.
			break;
		} else {
			/* end tp specified next token start address, therefore copy count is minus 1. */
			cp_cnt = end_tp - start_tp;
			strncpy(srv_conf_str, start_tp, cp_cnt);
			srv_conf_str[cp_cnt] = '\0';
			add_sconf(srv_conf_str, shead);
			total_cp_cnt += cp_cnt;
			memset(srv_conf_str, 0, MAX_LEN_RETURNED_SRV_CONFS);
		}
	}

	return 0;
}

int free_sconfq(struct server_confs_head * shead)
{
	struct server_confs * sconf = NULL;

	while ((sconf = TAILQ_FIRST(shead)) != NULL) {
		TAILQ_REMOVE(shead, sconf, entry);
		free(sconf);
		sconf = NULL;
	}

	return 0;
}

void print_sconfq(struct server_confs_head * shead)
{
	struct server_confs * sconf = NULL;

	for (sconf = TAILQ_FIRST(shead); sconf != NULL; sconf = TAILQ_NEXT(sconf, entry)) {
		printf("srv_addr: %s\n", sconf->srv_addr);
		printf("srv_port: %s\n", sconf->srv_port);
		printf("iface: %s\n", sconf->iface);
		printf("netmask: %s\n", sconf->netmask);
		printf("fakemac: %s\n", sconf->fakemac);
		printf("ri: %s\n", sconf->ri);
		printf("ri_opt: %s\n", sconf->ri_opt);
	}
}

