#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <string.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/queue.h>
#include <netinet/in.h>
#include <net/if.h>
#include <errno.h>

#include "common.h"
#include "dtsd_common.h"
#include "dtsd_util.h"
#include "dtsd_optparser.h"
#include "dtsd_server_conf.h"
#include "dtsd_command_exec.h"
#include "dtsd_worker.h"
#include "dtsd_ri_plugins.h"
#include "dtsd_plugin_loader.h"

dstatic int add_socket(int port, int socket);
dstatic int fetch_socket(const int port, int * socket);
static int date(int * status, char *, char *);
dstatic int get_macaddr(const char * iface, int * status, char *, char *);
dstatic int exec_add_vnodes(struct dtsd_node_info *, vnode_idset, char *, char *);
dstatic int set_ServerConfsStr(const struct server_conf *, const struct options *, int *, char *, char *);

dstatic int exec_ADD_CONF(struct dtsd_config *, struct options *, int *, char *, char *);
dstatic int exec_ADD_NODE(struct dtsd_config *, struct options *, int *, char *, char *);
dstatic int exec_DEL_NODE(struct dtsd_config *, struct options *, int *, char *, char *);

dstatic int dtsctl_join(char *, char *, char *, char *, int *, char *, char *);
dstatic int dtsctl_addnode(char *, char *, char *, char *, int *, char *, char *);
dstatic int dtsctl_removenode(char *, char *, char *, char *, int *, char *, char *);
dstatic int dtsctl_flush(int *, char *, char *);
dstatic int ifconfig_down(char *, int *, char *, char *);
dstatic int ifconfig_up(char *, int *, char *, char *);
dstatic int ifconfig_setup(char *, char *, char *, char *, int *, char *, char *);
dstatic int fakemac_setup(char *, int, int *, char *, char *);

void dtsd_command_exec(struct dtsd_config * dtsd_conf, const char * cmd, char * dtsd_res)
{
	struct options opt;
	char res[MAX_CMD_RES_SIZE] = "", err[MAX_CMD_ERR_SIZE] = "";
	int status = -1, ret = 0;

	debug_printf(HERE "about to execute command \"%s\".\n", cmd);

	if(dtsd_conf == NULL) {
		eprintf(HERE "dtsd_confing not specified.\n");
		return;
	}
	if(cmd == NULL) {
		eprintf(HERE "command string not specified.\n");
		return;
	}

	str2opt((char*)cmd, &opt, INNER_OPT_PARSE);

	if(get_mynode_state(dtsd_conf) != 1) {
		if(opt.cmd != CLEAR && opt.cmd != RESTORE) {
			debug_printf(HERE "dtsd don't synchronize a cluster configuration yet.\n");
			debug_printf(HERE "to synchronize the configuration, do Clear or Restore.\n");
			return;
		}
	}

	switch(opt.cmd) {
	case ADD_CONF: {
		ret = exec_ADD_CONF(dtsd_conf, &opt, &status, res, err);
	} break;
	case ADD_NODE: {
		ret = exec_ADD_NODE(dtsd_conf, &opt, &status, res, err);
	} break;
	case DEL_NODE: {
		ret = exec_DEL_NODE(dtsd_conf, &opt, &status, res, err);
	} break;
	case CLEAR: {
		struct server_conf * i;

		set_mynode_state(dtsd_conf, 1);
		// stop the all ri.
		for_each_server_conf(i, dtsd_conf->server_conf){
			if (i->ri_intf != NULL && i->ri != NULL) {
				if (i->ri_intf->stop(i->ri) != 0) {
					eprintf(HERE "failed to stop ri.\n");
				}
				if (i->ri_intf->finalize(i->ri) != 0) {
					eprintf(HERE "failed to finalize ri.\n");
				}
			}
		}

		free_server_conf_list(dtsd_conf->server_conf);
		dtsd_conf->server_conf = NULL;

		if (dtsctl_flush(&status, res, err) != 0) {
			status = -1;
			break;
		}
		debug_printf(HERE "succeeded to dtsctl flush.\n");

		set_mynode_state(dtsd_conf, 1);
		send_hbc_to_cluster();
		status = 0;

		break;
	}
	case LIST:
	case SAVE: {
		str2opt((char*)cmd, &opt, INNER_OPT_PARSE);
		set_ServerConfsStr(dtsd_conf->server_conf, &opt, &status, res, err);
		break;
	}
	case RESTORE: {
	}
	/* this command returns the candidate proxy port to next bind.    */
	/* if no proxy port still bound, then return the 0 as proxy port. */
	case GET_PROXY_PORT: {
		status = -1;
		debug_printf(HERE "accept " OPT_GET_PROXY_PORT" command.\n");

		if (dtsd_conf->last_proxy_port == 0) {
			sprintf(res, "%d", dtsd_conf->last_proxy_port);
		} else {
			sprintf(res, "%d", (dtsd_conf->last_proxy_port) + 1);
		}
		status = 0;
	} break;
	case BIND_SOCK: {
		int port = opt.proxy_port;
		status = -1;
		debug_printf(HERE "accept " OPT_BIND_SOCK " command.\n");

		dtsd_conf->last_proxy_port = port;
		if(bind_socket(port) != 0) {
			eprintf(HERE "failed to bind socket with port %d.", port);
		} else {
			debug_printf(HERE "succeeded to bind socket with port %d.", port);
			status = 0;
		}
	} break;
	case UNBIND_SOCK: {
		int port = opt.proxy_port;
		int socket = -1;
		status = -1;
		debug_printf(HERE "accept " OPT_BIND_SOCK " command.\n");
		if(fetch_socket(port, &socket) != 0) {
			eprintf(HERE "failed to bind socket with port %d.", port);
		} else {
			debug_printf(HERE "succeeded to bind socket with port %d.", port);
			close(socket);
			status = 0;
		}
	} break;
	case SET_MACADDR: {
		debug_printf(HERE "accept " OPT_SET_MACADDR " command.\n");
		debug_printf(HERE "detecting MAC address of interface \"%s\".\n", opt.iface);
		get_macaddr(opt.iface, &status, res, err);
	} break;
	case NONE:
#ifdef DEBUG
	case PRINTCONF: {
		char str[DTSD_MESSAGE_MAX];
		ServerConf2str(dtsd_conf->server_conf, str, sizeof(str));
		debug_printf("\n====== showing current confiugration =====\n");
		debug_printf(str);
		debug_printf("\n==========================================\n");
		status = 0;
	} break;
	case SHOWDATE: {
		debug_printf(HERE "dtsd: accept " OPT_SHOW_DATE " command.\n");
		date(&status, res, err);
	} break;
#endif
	default:
		break;
	}

	
	if(ret != 0 && (opt.cmd == ADD_NODE
					|| opt.cmd == ADD_CONF
					|| opt.cmd == DEL_NODE)) {
		int st;
		set_mynode_state(dtsd_conf, 0);
		dtsctl_flush(&st, NULL, NULL);
		free_server_conf_list(dtsd_conf->server_conf);
		dtsd_conf->server_conf = NULL;
		send_hbc_to_cluster();
	}

	sprintf(dtsd_res, ">>\nHOST-%s\nRESULT-%s\nERR-%s\nMSG-%s\n",
			dtsd_conf->hostname,
			status == 0?"OK":"NG",
			err, res);

	debug_printf(HERE "dtsd: command result is \"%s\"", dtsd_res);

	return;
}

struct ri * initialize_ri(int sock,
				const struct dtsd_config * dconf,
				const struct server_conf * sconf)
{
	return sconf
		->ri_intf
		->initialize(sock,
				dconf->naddr,
				sconf->server_port,
				sconf->ri_opt);
}

int ri_intf_init_run(struct dtsd_config * dtsd_conf,
			struct server_conf * sconf)
{
	int sock;

	if (sconf == NULL) {
		return -1;
	}

	// run the request inspector.
	if(fetch_socket(sconf->proxy_port, &sock) == -1) {
		eprintf(HERE "failed to fetch socket.\n");
		return -1;
	}

	debug_printf(HERE "succeeded to fetch socket.\n");


	sconf->ri_intf = find_ri_intf(sconf->riname);
	if(sconf->ri_intf != NULL) {
		sconf->ri = initialize_ri(sock, dtsd_conf, sconf);
		if(sconf->ri == NULL) {
			eprintf(HERE "failed to initialize Request Inspector.");
			return -1;
		}
		if(sconf->ri_intf->run(sconf->ri) != 0) {
			eprintf(HERE "failed to run Request Inspector.\n");
			return -1;
		}
		debug_printf(HERE "succeeded to run Request Inspector.\n");
	} else {
		return -1;
	}
	return 0;
}

int exec_ADD_CONF(struct dtsd_config * dtsd_conf,
				  struct options * opt,
				  int * status, char * res, char * err)
{
	struct server_conf * sconf;
	struct dtsd_node_info * ninfo;
	*status = -1;

	if(find_ServerConf(dtsd_conf, opt->host_srv) != NULL) {
		eprintf(HERE "Server address and port are duplicated.\n");
		strcpy(err, "Server address and port are duplicated.");
		return *status;
	}
	debug_printf(HERE "Adding new Server Confinguration.\n");

	sconf = (struct server_conf *) malloc (sizeof(struct server_conf));
	if(sconf == NULL) {
		eprintf(HERE "No Memory Available.\n");
		strcpy(err, "No Memory Avalilable.");
		return *status;
	}
	memset(sconf, 0, sizeof(struct server_conf));

	if(opt2ServerConf(opt, sconf) != 0 ){
		eprintf(HERE "failed to convert option to server configuration.\n");
		strcpy(err, "failed to convert option to server configuration.");
		free_server_conf_list(sconf);
		return *status;
	}
	debug_printf(HERE "succeeded to convert option to server configuration.\n");

	/* validate interface value. */
	if (strcmp(dtsd_conf->interface, opt->iface) == 0) {
		eprintf(HERE "inner interface specified: %s\n", opt->iface);
		strcpy(err, "inner interface specified.");
		free_server_conf_list(sconf);
		return *status;
	}

	// update the server configuration.
	if(add_ServerConf(dtsd_conf, sconf) != 0) {
		eprintf(HERE "failed to update Server Configuration.\n");
		free_server_conf_list(sconf);
		strcpy(err, "failed to update Server Configuration.");
		return *status;
	}
	debug_printf(HERE "succeeded to update Server Configuration.\n");

	if (exec_ServerConf(sconf, dtsd_conf, res, err) != 0) {
		debug_printf(HERE "failed to exec Server Configuration.\n");
		strcpy(err, "failed to exec Server Configuration.");
		return *status;
	}
	debug_printf(HERE "succeeded to apply Server Configuration.\n");

	// set the ri_intf.
	if (ri_intf_init_run(dtsd_conf, sconf) != 0) {
		*status = -1;
		debug_printf(HERE "failed to initialize and run the Request Inspector.\n");
		strcpy(err, "failed to initialize and run the Request Inspector.\n");
		return *status;
	}

	for_each_node_info(ninfo, sconf) {
		if (exec_NodeInfo(ninfo, sconf, res, err) != 0) {
			debug_printf(HERE "failed to exec Node Information.\n");
			strcpy(err, "failed to exec Node Information.");
			return *status;
		}
	}
	debug_printf(HERE "succeeded to apply Node Configuration.\n");

	*status = 0;
	return *status;
}

int exec_ADD_NODE(struct dtsd_config * dconf, struct options * opt,
				  int * status, char * res, char * err)
{
	struct server_conf * sconf;
	struct dtsd_node_info * ninfo, * i, * cur_ninfo = NULL;
	*status = -1;

	if((sconf = find_ServerConf(dconf, opt->host_srv)) == NULL) {
		eprintf(HERE "Server address and port are not configured.\n");
		return *status;
	}

	ninfo = (struct dtsd_node_info *) malloc (sizeof(struct dtsd_node_info));
	if(ninfo == NULL) {
		eprintf(HERE "No Memory Avalilable.\n");
		strcpy(err, "No Memory Avalilable.");
		return *status;
	}
	memset(ninfo, 0, sizeof(struct dtsd_node_info));
	debug_printf(HERE "succeeded to apply Node Info.\n");

	if(opt2NodeInfo(opt, ninfo) != 0) {
		eprintf(HERE "failed to convert options to node infomation.\n");
		strcpy(err, "failed to convert options to node infomation.");
		free_node_info(ninfo);
		return *status;
	}
	debug_printf(HERE "succeeded to convert options to node infomation.\n");

	for_each_node_info(i, sconf) {
		if(strcmp(i->hb_node_name, ninfo->hb_node_name) == 0) {
			cur_ninfo = i;
			break;
		}
	}

	if(cur_ninfo != NULL) {
		vnode_idset new_vnodes;
		new_vnodes = monus_IDset(ninfo->idset, cur_ninfo->idset);
		if(new_vnodes != 0 && cur_ninfo->hbc_node->ready == 1) {
			exec_add_vnodes(cur_ninfo, new_vnodes, res, err);
			debug_printf(HERE "succeeded to add virtual nodes.\n");
		} else {
			cur_ninfo->idset = ninfo->idset;
			debug_printf(HERE "virtual nodes are already exist or node not ready.\n");
			strcpy(err, "virtual nodes are already exist or node not ready.");
		}
		free_node_info(ninfo);

	} else {
		if(add_NodeInfo(sconf, ninfo) != 0) {
			eprintf(HERE "failed to add node infomation.\n");
			free_node_info(ninfo);
		}
		debug_printf(HERE "succeeded to add node infomation.\n");
		if (ninfo->hbc_node->ready == 0) {
			debug_printf(HERE "node %s is not readied yet\n", ninfo->hb_node_name);
		} else {
			if (exec_NodeInfo(ninfo, sconf, res, err) != 0) {
				debug_printf(HERE "failed to exec Node Information.\n");
				strcpy(err, "failed to exec Node Information.");
				return *status;
			}
			debug_printf(HERE "succeeded to exec Node Configuration.\n");
		}
	}

	*status = 0;
	return *status;
}

int disable_hbc_node(const struct server_conf * conf,
					 uint32_t addr,
					 char * res, char *err)
{
	char outer_addr_str[MAX_LEN_SRV_ADDR], intra_addr_str[MAX_LEN_SRV_ADDR];
	char server_port_str[MAX_LEN_SRV_PORT];
	struct in_addr in_addr;
	int status = -1;

	/* convert the outer_addr to addr string. */
	in_addr.s_addr = conf->server_address;
	strcpy(outer_addr_str, inet_ntoa(in_addr));

	/* convert the intra_addr to addr string */
	in_addr.s_addr = addr;
	strcpy(intra_addr_str, inet_ntoa(in_addr));

	/* convert the server_port to port string. */
	snprintf(server_port_str, MAX_LEN_SRV_PORT, "%d", conf->server_port);
	
	if (dtsctl_removenode(outer_addr_str, server_port_str, intra_addr_str, "0",
						  &status, res, err) != 0) {
		eprintf(HERE "failed to dtsctl removenode.\n");
		// status, res, err is set by dtsctl_removenode().
		return status;
	}
	return 0;
}

int exec_remove_vnodes(struct server_conf * conf,
					   const char * hbc_node_name,
					   vnode_idset tobe_removed,
					   int * status, char * res, char * err)
{
	struct dtsd_node_info * i = NULL, * cur_info = NULL;
	struct hbc_node * hbc_node;
	char outer_addr_str[MAX_LEN_SRV_ADDR], intra_addr_str[MAX_LEN_SRV_ADDR];
	char server_port_str[MAX_LEN_SRV_PORT];
	struct in_addr in_addr;
	int id;

	for_each_node_info(i, conf) {
		if(strcmp(i->hb_node_name, hbc_node_name) == 0) {
			cur_info = i;
			break;
		}
	}

	if(cur_info == NULL) {
		eprintf(HERE "no Node Configuration named \"%s\".\n", hbc_node_name);
		return *status;
	}

	cur_info->idset = monus_IDset(cur_info->idset, tobe_removed);
	debug_printf(HERE "update Node Information.\n");

	hbc_node = find_hbc_node(hbc_node_name);
	if (hbc_node == NULL || hbc_node->ready != 1) {
		debug_printf(HERE "Heartbeat Client %s is not active\n", hbc_node_name);
		debug_printf(HERE "succeeded to remove node infomation.\n");
		*status = 0;
		return *status;
	}

	in_addr.s_addr = conf->server_address;
	strcpy(outer_addr_str, inet_ntoa(in_addr));
	snprintf(server_port_str, MAX_LEN_SRV_PORT, "%d", conf->server_port);
	in_addr.s_addr = hbc_node->addr;
	strcpy(intra_addr_str, inet_ntoa(in_addr));

	for_each_idset(id, tobe_removed) {
		int len;
		char id_str[MAX_LEN_VNODE_ID];
		len = sprintf(id_str, "%d", id);
		if (dtsctl_removenode(outer_addr_str, server_port_str,
							  intra_addr_str, id_str,
							  status, res, err) != 0) {
			eprintf(HERE "failed to dtsctl add.\n");
			return -1;
		}
	}
	debug_printf(HERE "succeeded to dtsctl add.\n");

	return 0;
}

int exec_remove_pnode(struct server_conf * conf,
					  const char * hbc_node_name,
					  int * status, char * res, char * err)
{
	struct dtsd_node_info * i = NULL, * ninfo = NULL;
	struct hbc_node * hbc_node;

	for_each_node_info(i, conf) {
		if(strcmp(i->hb_node_name, hbc_node_name) == 0) {
			ninfo = NULL;
			break;
		}
	}
	if(ninfo == NULL) {
		eprintf(HERE "no Node Configuration named \"%s\".\n", hbc_node_name);
	}

	if (remove_NodeInfo(conf, hbc_node_name) != 0) {
		eprintf(HERE "failed to remove node information.\n");
		return *status;
	}

	hbc_node = find_hbc_node(hbc_node_name);
	if (hbc_node == NULL || hbc_node->ready != 1) {
		debug_printf(HERE "Heartbeat Client %s is not active\n", hbc_node_name);
		debug_printf(HERE "succeeded to remove node infomation.\n");
		*status = 0;
		return *status;
	}

	if (disable_hbc_node(conf, hbc_node->addr, res, err) != 0) {
		return *status;
	}

	if (conf->ri_intf == NULL) {
		eprintf(HERE "No Requet Interface Running.\n");
	} else {
		if (conf->ri_intf->remove_node(conf->ri, hbc_node->addr) != 0) {
			eprintf(HERE "failed to remove node from Requet Interface.\n");
			return *status;
		}
		eprintf(HERE "succeeded to remove node from Requet Interface.\n");
	}

	return 0;
}

int exec_DEL_NODE(struct dtsd_config * dconf, struct options * opt,
				  int * status, char * res, char * err)
{
	struct server_conf * sconf;
	char * nodespec;
	char * hbc_node_name;
	vnode_idset idset;

	*status = -1;

	if((sconf = find_ServerConf(dconf, opt->host_srv)) == NULL) {
		eprintf(HERE "Server address and port are not configured.\n");
		return *status;
	}

	nodespec = opt->nodes;
	if(strchr(nodespec, '/') == NULL) {
		hbc_node_name = nodespec;
		idset = 0;
	} else {
		if(parse_nodespec(nodespec, &hbc_node_name, &idset) == -1) {
			debug_printf(HERE "failed to parse node spec.");
			return *status;
		}
	}

	if(idset == 0 || contain_IDset(0, idset)) {
		debug_printf(HERE "about to remove pnode named \"%s\"\n", hbc_node_name);
		if(exec_remove_pnode(sconf, hbc_node_name, status, res, err) != 0) {
			return *status;
		}
		debug_printf(HERE "succeeded to remove node.\n");
	} else {
		debug_printf(HERE "about to remove vnode node of \"%s\"\n", hbc_node_name);
		if(exec_remove_vnodes(sconf, hbc_node_name, idset, status, res, err) != 0) {
			return *status;
		}
		debug_printf(HERE "succeeded to remove vnodes.\n");
	}

	*status = 0;
	return *status;
}

int exec_ServerConf(const struct server_conf * server_conf,
					const struct dtsd_config * dtsd_conf,
					char * res, char * err)
{
	int status;
	char outer_addr_str[MAX_LEN_SRV_ADDR], outer_netmask_str[MAX_LEN_NETMASK];
	char server_port_str[MAX_LEN_SRV_PORT], proxy_port_str[MAX_LEN_PROXY_PORT];
	struct in_addr in_addr;
	// down the iface for set the macaddr.
	if (ifconfig_down(server_conf->iface, &status, res, err) != 0) {
		eprintf(HERE "failed to down the network interface\n");
		return -1;
	}
	debug_printf(HERE "succeeded to down the network interface.\n");

	in_addr.s_addr = server_conf->server_address;
	strcpy(outer_addr_str, inet_ntoa(in_addr));

	in_addr.s_addr = server_conf->netmask;
	strcpy(outer_netmask_str, inet_ntoa(in_addr));

	// set the ipaddress, netmask, macaddr to iface.
	if (ifconfig_setup(server_conf->iface, outer_addr_str,
				outer_netmask_str, server_conf->hwaddr,
				&status, res, err) != 0) {
		eprintf(HERE "failed to setup the network interface\n");
		return -1;
	}
	debug_printf(HERE "succeeded to setup the network interface\n");

	// up the iface.
	if (ifconfig_up(server_conf->iface, &status, res, err) != 0) {
		eprintf(HERE "failed to up the network interface.\n");
		return -1;
	}
	debug_printf(HERE "succeeded to up the network interface.\n");

	// set the fakemac if spcefied.
	if (fakemac_setup(server_conf->iface, server_conf->fakemac, &status, res, err) != 0) {
		eprintf(HERE "failed to setup the fakemac.\n");
		return -1;
	}
	debug_printf(HERE "succeeded to setup the fakemac.\n");

	snprintf(server_port_str, MAX_LEN_SRV_PORT, "%d", server_conf->server_port);
	snprintf(proxy_port_str, MAX_LEN_PROXY_PORT, "%d", server_conf->proxy_port);
	if (dtsctl_join(outer_addr_str, dtsd_conf->ipaddr, server_port_str,
				proxy_port_str, &status, res, err) != 0) {
		eprintf(HERE "failed to dtsctl join.\n");
		return -1;
	}
	debug_printf(HERE "succeeded to dtsctl join.\n");

	return 0;
}

int exec_add_vnodes(struct dtsd_node_info * info,
					vnode_idset new_vnodes,
					char * res, char * err)
{
	int status, id, len;
	char outer_addr_str[MAX_LEN_SRV_ADDR], intra_addr_str[MAX_LEN_SRV_ADDR];
	char server_port_str[MAX_LEN_SRV_PORT];
	char id_str[MAX_LEN_VNODE_ID];
	struct server_conf * sconf = info->srv_conf;
	struct in_addr in_addr;

	/* convert the outer_addr to addr string. */
	in_addr.s_addr = sconf->server_address;
	strcpy(outer_addr_str, inet_ntoa(in_addr));
	/* convert the server_port to port string. */
	snprintf(server_port_str, MAX_LEN_SRV_PORT, "%d", sconf->server_port);

	/* convert the intra_addr to addr string */
	in_addr.s_addr = info->hbc_node->addr;
	strcpy(intra_addr_str, inet_ntoa(in_addr));

	// add the node to L4 load balancing.
	for_each_idset(id, new_vnodes) {
		len = sprintf(id_str, "%d", id);
		if (dtsctl_addnode(outer_addr_str, server_port_str,
					intra_addr_str, id_str,
					&status, res, err) != 0) {
			eprintf(HERE "failed to dtsctl add.\n");
			return -1;
		}
	}
	debug_printf(HERE "succeeded to dtsctl add.\n");
	// add the node to L7 load balancing.
	if(info->idset == 0) {
		if (sconf->ri_intf == NULL) {
			eprintf(HERE "request inspector is NULL.\n");
			return -1;
		}
		if (sconf->ri_intf->add_node(sconf->ri, info->hbc_node->addr) != 0) {
			eprintf(HERE "failed to ri_intf->add_node.\n");
			return -1;
		}
		debug_printf(HERE "succeeded to ri_intf->add_node.\n");
	}

	info->idset = add_IDset(info->idset, new_vnodes);
	return 0;
}

int exec_NodeInfo(const struct dtsd_node_info * info,
				  const struct server_conf * server_conf,
				  char * res, char * err)
{
	int status, id, len;
	char outer_addr_str[MAX_LEN_SRV_ADDR], intra_addr_str[MAX_LEN_SRV_ADDR];
	char server_port_str[MAX_LEN_SRV_PORT];
	char id_str[MAX_LEN_VNODE_ID];
	struct in_addr in_addr;

	/* if hbc_node not set, then can not exec. */
	if(info->hbc_node == NULL) {
		eprintf(HERE "dtsd_node_info->hbc_node is not set. hb_node_name: %s\n", info->hb_node_name);
		return 0;
	}

	if(info->hbc_node->ready != 1) {
		debug_printf(HERE "node %s is not readied yet.\n", info->hb_node_name);
		// not error
		return 0;
	}

	/* convert the outer_addr to addr string. */
	in_addr.s_addr = server_conf->server_address;
	strcpy(outer_addr_str, inet_ntoa(in_addr));
	/* convert the server_port to port string. */
	snprintf(server_port_str, MAX_LEN_SRV_PORT, "%d", server_conf->server_port);

	/* convert the intra_addr to addr string */
	in_addr.s_addr = info->hbc_node->addr;
	strcpy(intra_addr_str, inet_ntoa(in_addr));

	// add the node to L4 load balancing.
	if (info->idset == 0) {
		if (dtsctl_addnode(outer_addr_str, server_port_str,
					intra_addr_str, "0",
					&status, res, err) != 0) {
			eprintf(HERE "failed to dtsctl add.\n");
			return -1;
		}
	} else {
		for_each_idset(id, info->idset) {
			len = sprintf(id_str, "%d", id);
			if (dtsctl_addnode(outer_addr_str, server_port_str,
						intra_addr_str, id_str,
						&status, res, err) != 0) {
				eprintf(HERE "failed to dtsctl add.\n");
				return -1;
			}
		}
	}
	debug_printf(HERE "succeeded to dtsctl add.\n");
	// add the node to L7 load balancing.
	if (server_conf->ri_intf == NULL) {
		eprintf(HERE "Request Inspector is NULL.\n");
		return -1;
	} else {
		if (server_conf->ri_intf->add_node(server_conf->ri, info->hbc_node->addr) != 0) {
			eprintf(HERE "failed to add node to Request Inspector.\n");
			return -1;
		}
		debug_printf(HERE "succeeded to add node to Request Inspector.\n");
	}

	return 0;
}

int exec_cmd(char * const args[], int * status, char * res, char * err)
{
	pid_t cpid;
	int stdout_pipefd[2];
	int stderr_pipefd[2];

	if (pipe(stdout_pipefd) == -1) {
		perror("pipe");
		return -1;
	}
	if (pipe(stderr_pipefd) == -1) {
		perror("pipe");
		return -1;
	}

	if((cpid = fork()) == -1){
		perror("fork");
		exit(-1);
	}

	if(cpid == 0) {
		if(close(stdout_pipefd[0]) < 0) {
			perror("close");
			exit(1);
		}
		if(close(stderr_pipefd[0]) < 0) {
			perror("close");
			exit(1);
		}
		if(dup2(stdout_pipefd[1], STDOUT_FILENO) < 0) {
			perror("dup2");
			exit(1);
		}
		if(dup2(stderr_pipefd[1], STDERR_FILENO) < 0) {
			perror("dup2");
			exit(1);
		}
		if(close(stdout_pipefd[1]) < 0) {
			perror("close");
			exit(1);
		}
		if(close(stderr_pipefd[1]) < 0) {
			perror("close");
			exit(1);
		}
		if(execv(args[0], &args[0]) == -1) {
			perror("execv");
			exit(0);
		}
		return 0; // never reached.

	} else {
		int ret, st, size;

		close(stdout_pipefd[1]);
		close(stderr_pipefd[1]);

		if((ret = waitpid(cpid, &st, WUNTRACED)) != cpid) {
			eprintf("dtsd: some error in child process.\n");
		}

		*status = WEXITSTATUS(st);

		if(res == NULL && err == NULL)
			goto close_and_return;

		if(res != NULL) {
			size = MAX_CMD_RES_SIZE;
			if((ret = read(stdout_pipefd[0], res, size)) < 0) {
				perror("read");
				exit(1);
			}
			if(ret >= size) {
				// too long
				res[ret-1] = '\0';
				*status = -1;
				strcpy(err, "dtsd: command output is too large!");
				goto close_and_return;
			} else {
				res[ret] = '\0';
			}
		}

		if(err == NULL)
			goto close_and_return;

		if((ret = read(stderr_pipefd[0], err, size)) < 0) {
			perror("read");
			exit(1);
		}
		if(ret >= size)
			err[ret-1] = '\0';
		else
			err[ret] = '\0';

close_and_return:
		close(stdout_pipefd[0]);
		close(stderr_pipefd[0]);

		return 0;
	}
}

int date(int * status, char * res, char * err)
{
	exec_cmd((char * []){"/bin/date", (char *)NULL}, status, res, err);
	return 0;
}

#if DEBUG
dstatic int hello_world(int * status, char * res, char * err)
{
	exec_cmd((char * []){"/bin/echo", "Hello World", NULL}, status, res, err);
	debug_fprintf(stderr, "stdout = %s\n", *res);
	debug_fprintf(stderr, "stderr = %s\n", *err);
	return 0;
}
#endif

int dtsctl_join(char * outer_address, char * inner_address,
		char * outer_port, char * proxy_port,
		int * status, char * res, char * err)
{
	exec_cmd((char * []) {DTSCTL_PATH "dtsctl",
					"join",
					outer_address,
					inner_address,
					outer_port,
					proxy_port,
					NULL}, status, res, err);
	return *status;
}

int dtsctl_flush(int * status, char * res, char * err)
{
	exec_cmd((char * [])  {DTSCTL_PATH "dtsctl",
					 "flush",
					 NULL}, status, res, err);
	return *status;
}

int dtsctl_list(int * status, char * res, char * err)
{
	exec_cmd((char * []) {DTSCTL_PATH "dtsctl",
					"flush",
					NULL}, status, res, err);
	return * status;
}

int dtsctl_addnode(char * outer_address, char * outer_port,
			char * intra_address, char * vnode_id,
			int * status, char * res, char * err)
{
	if (strlen(vnode_id) == 0) {
		exec_cmd((char * []) {DTSCTL_PATH "dtsctl",
						"addnode",
						outer_address,
						outer_port,
						intra_address,
						NULL}, status, res, err);
	} else {
		exec_cmd((char * []) {DTSCTL_PATH "dtsctl",
						"addnode",
						outer_address,
						outer_port,
						intra_address,
						vnode_id,
						NULL}, status, res, err);
	}

	return *status;
}

int dtsctl_removenode(char * outer_address, char * outer_port, char * intra_address,
					  char * id, int * status, char * res, char * err)
{
	exec_cmd((char * []) {DTSCTL_PATH "dtsctl",
				"removenode",
				outer_address,
				outer_port,
				intra_address,
				id,
				NULL}, status, res, err);
	return *status;
}

int ifconfig_down(char * iface, int * status, char * res, char * err)
{
	exec_cmd((char * []) {IFCONFIG_PATH "ifconfig",
					iface,
					"down",
					NULL}, status, res, err);
	return *status;
}

int ifconfig_up(char * iface, int * status, char * res, char * err)
{
	exec_cmd((char * []) {IFCONFIG_PATH "ifconfig",
					iface,
					"up",
					NULL}, status, res, err);
	return *status;
}

int ifconfig_setup(char * iface, char * outer_address, char * netmask, char * macaddr,
			int * status, char * res, char * err)
{
	exec_cmd((char * []) {IFCONFIG_PATH "ifconfig",
					iface,
					outer_address,
					"netmask",
					netmask,
					"hw",
					"ether",
					macaddr,
					NULL}, status, res, err);
	return *status;
}

int fakemac_setup(char * iface, int fakemac_stat, int * status, char * res, char * err)
{
	char cmd[WORK_BUF_SIZE] = "";

	snprintf(cmd, WORK_BUF_SIZE, "%s/echo %s > %s/%s", ECHO_PATH,
							fakemac_stat == 1 ? "1" : "0",
							FAKEMAC_PATH,
							iface);
	*status = system(cmd);
	return *status;
}

int get_macaddr(const char * iface, int * status, char * res, char * err)
{
	int sock;
	struct ifreq ifr;
	*status = -1;

	if((sock = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
		perror("socket");
		goto exception;
	}
	ifr.ifr_addr.sa_family = AF_INET;
	strncpy(ifr.ifr_name, iface, IFNAMSIZ - 1);
	if((ioctl(sock, SIOCGIFHWADDR, &ifr)) == -1) {
		perror("ioctl");
		goto exception;
	}
	close(sock);

	snprintf(res, MAX_LEN_MACADDR, "%02x:%02x:%02x:%02x:%02x:%02x",
			 (unsigned char)ifr.ifr_hwaddr.sa_data[0],
			 (unsigned char)ifr.ifr_hwaddr.sa_data[1],
			 (unsigned char)ifr.ifr_hwaddr.sa_data[2],
			 (unsigned char)ifr.ifr_hwaddr.sa_data[3],
			 (unsigned char)ifr.ifr_hwaddr.sa_data[4],
			 (unsigned char)ifr.ifr_hwaddr.sa_data[5]);

	debug_printf(HERE "MAC address of %s is %s\n", iface, res);

	*status = 0;
	return *status;
exception:
	res = "";
	strcpy(err, "failed to configure macaddr.");
	return *status;
}


struct socket_container {
	struct socket_container * next;
	int socket;
	int port;
};

struct {
	struct socket_container * next;
	struct socket_container * last;
}
SOCKET_CONTAINER = {
	.next = NULL,
	.last = NULL,
};

int add_socket(int port, int socket)
{
	struct socket_container * new;
	new = (struct socket_container *) malloc (sizeof(struct socket_container));
	if(new == NULL) {
		eprintf(HERE "No memory available.\n");
		return -1;
	}
	memset(new, 0, sizeof(struct socket_container));
	new->next = NULL;
	new->socket = socket;
	new->port = port;
	if(SOCKET_CONTAINER.next == NULL) {
		SOCKET_CONTAINER.next = new;
	} else {
		SOCKET_CONTAINER.last->next = new;
	}
	SOCKET_CONTAINER.last = new;
	return 0;
}

int fetch_socket(const int port, int * socket)
{
	struct socket_container * cont, * prev = NULL;
	for(cont = SOCKET_CONTAINER.next; cont != NULL; prev = cont, cont = cont->next) {
		if(cont->port == port) {
			if(prev == NULL) {
				SOCKET_CONTAINER.next = cont->next;
			} else {
				prev->next = cont->next;
			}
			if(SOCKET_CONTAINER.last == cont) {
				SOCKET_CONTAINER.last = prev;
			}
			*socket = cont->socket;
			free(cont);
			return 0;
		}
	}
	eprintf(HERE "No bound socket with port %d\n", port);
	return -1;
}

int bind_socket(const int port)
{
	int sock;
	struct sockaddr_in sin;
	int reuse = 1;
	int e;

	sin.sin_family      = AF_INET;
	sin.sin_addr.s_addr = INADDR_ANY;
	sin.sin_port        = htons(port);

	sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (sock < 0)
		return -errno;

	if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &reuse, sizeof(reuse)) < 0)
		goto out_close;

	if (bind(sock, (struct sockaddr *) &sin, sizeof(sin)) < 0)
		goto out_close;

	if(add_socket(port, sock) == -1)
		goto out_close;

	return 0;

out_close:
	e = errno;
	close(sock);
	return -e;
}

enum node_info_status {
	ALIVE_VNODE = 0,
	ONLY_PNODE,
	NOT_ACTIVE,
};

struct node_info {
	char hb_node_name[MAX_LEN_HOST];
	char addr[MAX_LEN_SRV_ADDR];
	char id[MAX_LEN_VNODE_ID];
	char id_hash[MAX_LEN_VNODE_HASH_ID];
	enum node_info_status status;

	TAILQ_ENTRY(node_info) entry;
};

TAILQ_HEAD(node_info_head, node_info);

static
void free_ninfoq(struct node_info_head * vihead)
{
	struct node_info * vi;

	while ((vi = TAILQ_FIRST(vihead)) != NULL) {
		TAILQ_REMOVE(vihead, vi, entry);
		free(vi);
		vi = NULL;
	}
}

/* create the node info queue for save command. */
static
struct node_info_head * create_ninfoq_for_save(const struct server_conf * conf)
{
	int id;
	struct dtsd_node_info * dtsd_ninfo;

	struct node_info * to_add;
	struct node_info_head * vihead;

	vihead = (struct node_info_head *) malloc(sizeof(struct node_info_head));
	if (vihead == NULL) {
		return NULL;
	}
	memset(vihead, 0, sizeof(struct node_info_head));
	TAILQ_INIT(vihead);

	for_each_node_info(dtsd_ninfo, conf) {
		char hbc_node_addr_str[MAX_LEN_SRV_ADDR];
		struct in_addr addr;

		addr.s_addr = dtsd_ninfo->hbc_node->addr;
		strcpy(hbc_node_addr_str, inet_ntoa(addr));

		/* no id exists. */
		if (dtsd_ninfo->idset == 0) {
			to_add = (struct node_info *) malloc(sizeof(struct node_info));
			if (to_add == NULL) {
				debug_printf(HERE "failed to allocate node_info.\n");
				goto free4fail;
			} else {
				memset(to_add, 0, sizeof(struct node_info));
			}

			strcpy(to_add->addr, inet_ntoa(addr));
			strcpy(to_add->hb_node_name, dtsd_ninfo->hb_node_name);
			strcpy(to_add->id, "0");
			to_add->status = ONLY_PNODE;

			/* add to the end of info. */
			TAILQ_INSERT_TAIL(vihead, to_add, entry);
			continue;
		}

		/* id exists. */
		for_each_idset(id, dtsd_ninfo->idset) {
			char id_str[MAX_LEN_VNODE_ID] = "";

			/* get vnode id hash value. */
			to_add = (struct node_info *) malloc(sizeof(struct node_info));
			if (to_add == NULL) {
				debug_printf(HERE "failed to allocate node_info.\n");
				goto free4fail;
			} else {
				memset(to_add, 0, sizeof(struct node_info));
			}

			sprintf(id_str, "%d", id);
			strcpy(to_add->addr, inet_ntoa(addr));
			strcpy(to_add->hb_node_name, dtsd_ninfo->hb_node_name);
			strcpy(to_add->id, id_str);
			to_add->status = ALIVE_VNODE;
			/* id hash not need for save the command. */

			TAILQ_INSERT_TAIL(vihead, to_add, entry);
		}
	}
	goto free4success;

free4fail:
	free_ninfoq(vihead);
	return NULL;
free4success:
	return vihead;
}

/* create the node info queue for list command. */
static
struct node_info_head * create_ninfoq_for_list(const struct dtsd_node_info * info,
						int dead_node,
						const char * outer_address,
						const char * outer_port)
{
	FILE * fp;
	in_addr_t addr_t;
	int cnt = 0;
	char node_file_path[WORK_BUF_SIZE] = "", line[WORK_BUF_SIZE];
	char * tp, * sp;
	struct node_info * to_add;
	struct node_info_head * vihead;
	const struct dtsd_node_info * work_info;
	int i, node_is_alive[MAX_HBC_NODES];

	for(i = 0; i < MAX_HBC_NODES; i++)
		node_is_alive[i] = 0;

	/* decide the file path. */
	if ((addr_t = inet_addr(outer_address)) == INADDR_NONE) {
		eprintf(HERE "failed to inet_addr.\n");
		return NULL;
	}
	sprintf(node_file_path, "%s%08X%04X",
			DTS_CONFIG_PATH, ntohl(addr_t), atoi(outer_port));
	debug_printf(HERE "get vnode from: %s\n", node_file_path);

	if ((fp = fopen(node_file_path, "r")) == NULL) {
		eprintf(HERE "failed to open node file.\n");
		return NULL;
	}

	vihead = (struct node_info_head *) malloc(sizeof(struct node_info_head));
	if (vihead == NULL) {
		eprintf(HERE "No Memory Avalilable.\n");
		return NULL;
	}
	memset(vihead, 0, sizeof(struct node_info_head));
	TAILQ_INIT(vihead);

	while(fgets(line, WORK_BUF_SIZE, fp) != NULL) {
		int i_addr = 0;
		struct in_addr addr;

		/* if cnt equals 0, the line is the header of node info. */
		if (cnt == 0) {
			cnt++;
			continue;
		}

		if (line[strlen(line) - 1] == '\n') {
			line[strlen(line) - 1] = '\0';
		}

		to_add = (struct node_info *) malloc(sizeof(struct node_info));
		if (to_add == NULL) {
			eprintf(HERE "No Memory Avalilable.\n");
			goto exception;
		} else {
			memset(to_add, 0, sizeof(struct node_info));
		}

		/* get the intra addr. */
		tp = strtok_r(line, " \t", &sp);
		if (tp == NULL) {
			free(to_add);
			continue;
		}
		sscanf(tp, "%x", &i_addr);
		addr.s_addr = htonl(i_addr);
		strcpy(to_add->addr, inet_ntoa(addr));

		/* get_the hb_node_name and check if the node info exist. */
		char hbc_node_addr_str[MAX_LEN_SRV_ADDR];
		int is_find_info = 0;
		i = -1;
		work_info = info;
		while (work_info != NULL) {
			i++;
			if (work_info->hbc_node == NULL) {
				eprintf(HERE "dtsd_node_info->hbc_node is not set. hb_node_name: %s\n", work_info->hb_node_name);
				continue;
			}
			addr.s_addr = work_info->hbc_node->addr;
			strcpy(hbc_node_addr_str, inet_ntoa(addr));
			if (strcmp(to_add->addr, hbc_node_addr_str) == 0) {
				strcpy(to_add->hb_node_name, work_info->hb_node_name);
				to_add->status = ALIVE_VNODE;
				node_is_alive[i] = 1;
				is_find_info = 1;
				break;
			}
			work_info = work_info->next;
		}
		if (is_find_info == 0) {
			eprintf(HERE "unknown node described at vnodes file: %s\n", to_add->addr);
			/* treat the addr as hb_node_name if node not found. */
			strcpy(to_add->hb_node_name, to_add->addr);
			to_add->status = ALIVE_VNODE;
		}

		/* get vnode id. */
		tp = strtok_r(NULL, " \t", &sp);
		if (tp == NULL) {
			free(to_add);
			continue;
		}
		strcpy(to_add->id, tp);

		/* get vnode id hash value. */
		tp = strtok_r(NULL, " \t", &sp);
		if (tp == NULL) {
			free(to_add);
			continue;
		}
		strcpy(to_add->id_hash, tp);

		/* add to the end of info. */
		TAILQ_INSERT_TAIL(vihead, to_add, entry);
	}

	i = 0;
	for(work_info = info; work_info != NULL; work_info = work_info->next) {
		if(node_is_alive[i] == 0) {
			to_add = (struct node_info *) malloc (sizeof(struct node_info));
			if (to_add == NULL) {
				eprintf(HERE "failed to allocate node_info.\n");
				goto exception;
			}
			memset(to_add, 0, sizeof(struct node_info));
			strcpy(to_add->hb_node_name, work_info->hb_node_name);
			if (work_info->hbc_node == NULL || work_info->hbc_node->ready != 1) {
				to_add->status = NOT_ACTIVE;
			} else {
				to_add->status = ONLY_PNODE;
			}
			if (dead_node == 0 && to_add->status == NOT_ACTIVE) {
				free(to_add);
				i++;
				continue;
			}
			TAILQ_INSERT_TAIL(vihead, to_add, entry);
		}
		i++;
	}

	fclose(fp);
	return vihead;

exception:
	if(vihead != NULL)
		free_ninfoq(vihead);
	fclose(fp);
	return NULL;
}

/* append the str to dst. */
static
int append_str(char * dst, char * str, int max_dst_len)
{
	if (dst == NULL || str == NULL) {
		eprintf(HERE "dst or str is NULL.\n");
		return -1;
	}
	if (max_dst_len <= 0) {
		eprintf(HERE "specified max dst len is <= 0.\n");
		return -1;
	}
	if (((int) strlen(dst) + (int) strlen(str)) > max_dst_len) {
		eprintf(HERE "nifno str len over the max len.\n");
		return -1;
	}
	strcat(dst, str);
	return 0;
}

/* create following format form vnodeq. */
/* dts01%10.0.0.1/1#00000001:dts02%10.0.0.2/2#00000002:dts03%10.0.0.3/#:dts04(dead)%/#: */
static
void ninfoq2str(struct node_info_head * vihead, char * vnodes_str, int max_str_len)
{
	struct node_info * vi;

	for (vi = TAILQ_FIRST(vihead); vi != NULL; vi = TAILQ_NEXT(vi, entry)) {
		if (append_str(vnodes_str, vi->hb_node_name, max_str_len) != 0) {
			return;
		}
		if (vi->status == NOT_ACTIVE) {
			if (append_str(vnodes_str, TOKEN_DEAD, max_str_len) != 0) {
				return;
			}
		}
		if (append_str(vnodes_str, TOKEN_PERCENT, max_str_len) != 0) {
			return;
		}
		if (append_str(vnodes_str, vi->addr, max_str_len) != 0) {
			return;
		}
		if (append_str(vnodes_str, TOKEN_SLASH, max_str_len) != 0) {
			return;
		}
		if (append_str(vnodes_str, vi->id, max_str_len) != 0) {
			return;
		}
		if (append_str(vnodes_str, TOKEN_SHARP, max_str_len) != 0) {
			return;
		}
		if (append_str(vnodes_str, vi->id_hash, max_str_len) != 0) {
			return;
		}
		if (append_str(vnodes_str, TOKEN_COLON, max_str_len) != 0) {
			return;
		}
	}
}

/* set the each server conf to res .                          */
/* dead_node(argument 2) is the flag, if its value is 1,      */
/* get the dead node configuration with alive one.            */
/* if value is 0, then get the only alive node configuration. */
dstatic
int set_ServerConfsStr(const struct server_conf * sconf,
			const struct options * opt,
			int * status,
			char * res,
			char * err)
{
	char work[WORK_BUF_SIZE] = "";
	const struct server_conf * i;

	for_each_server_conf(i, sconf){
		struct node_info_head * vihead = NULL;
		char outer_addr_str[MAX_LEN_SRV_ADDR] = "", outer_netmask_str[MAX_LEN_NETMASK] = "";
		char server_port_str[MAX_LEN_SRV_PORT] = "";
		char vnodes_str[MAX_LEN_RETURNED_NODES] = "";
		struct in_addr in_addr;

		/* convert server_address to string. */
		in_addr.s_addr = i->server_address;
		strcpy(outer_addr_str, inet_ntoa(in_addr));

		/* convert netmask string. */
		in_addr.s_addr = i->netmask;
		strcpy(outer_netmask_str, inet_ntoa(in_addr));

		sprintf(server_port_str, "%d", i->server_port);

		/* get vnode info queue. */
		if (opt->cmd == LIST) {
			vihead = create_ninfoq_for_list(i->ni, opt->dead_node, outer_addr_str, server_port_str);
		} else if (opt->cmd == SAVE) {
			vihead = create_ninfoq_for_save(i);
		}
		if (vihead == NULL) {
			continue;
		}

		ninfoq2str(vihead, vnodes_str, MAX_CMD_RES_SIZE);

		/*
		 * create the following format.
		 *
		 * ?srv_addr=192.168.30.1?srv_port=80?iface=eth0\
		 * ?hwaddr=aa:bb:cc:dd:ee:ff?netmask=255.255.255.0?fakemac=on\
		 * ?ri=httpri?ri_opt=rr\?proxy_port=53272\
		 * ?node=dts01%10.0.0.1/1#00000001:dts02%10.0.0.2/2#00000002:dts03:dts04(dead)
		 */
		sprintf(work, "%s%s%s%s%s%s%s%d%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%d%s%s%s%s",
					TOKEN_QUESTION, TOKEN_SRV_ADDR,   TOKEN_EQUAL, outer_addr_str,
					TOKEN_QUESTION, TOKEN_SRV_PORT,   TOKEN_EQUAL, i->server_port,
					TOKEN_QUESTION, TOKEN_IFACE,      TOKEN_EQUAL, i->iface,
					TOKEN_QUESTION, TOKEN_MACADDR,    TOKEN_EQUAL, i->hwaddr,
					TOKEN_QUESTION, TOKEN_NETMASK,    TOKEN_EQUAL, outer_netmask_str,
					TOKEN_QUESTION, TOKEN_FAKEMAC,    TOKEN_EQUAL, i->fakemac==1?"on":"off",
					TOKEN_QUESTION, TOKEN_RI,         TOKEN_EQUAL, i->riname,
					TOKEN_QUESTION, TOKEN_RI_OPT,     TOKEN_EQUAL, i->ri_opt,
					TOKEN_QUESTION, TOKEN_PROXY_PORT, TOKEN_EQUAL, i->proxy_port,
					TOKEN_QUESTION, TOKEN_NODES,      TOKEN_EQUAL, vnodes_str);

		if ((strlen(res) + strlen(work)) > MAX_CMD_RES_SIZE) {
			eprintf(HERE "over the response size.\n");
			strcpy(err, "over the response size.\n");
			*status = -1;
			return *status;
		}
		/* concatenate the each server conf str. */
		strcat(res, work);

		free_ninfoq(vihead);
	}

	*status = 0;
	return *status;
}

