#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <netdb.h>

#include "common.h"
#include "dtsd_common.h"
#include "dtsd_util.h"

int serialize_idset(vnode_idset idset, char * str)
{
	int id, len, count = 0;
	if(str == NULL) {
		return -1;
	}
	for_each_idset(id, idset) {
		if(count > 0) {
			*str = ',';
			str++;
		}
		len = sprintf(str, "%d", id);
		str += len;
		count++;
	}
	return 0;
}

int deserialize_idset(char * str, vnode_idset * idset)
{
	return parse_idset(str, idset);
}

int resolve_host(const char * hostname, uint32_t * addr)
{
	struct hostent *ent = gethostbyname(hostname);
	if (ent == NULL) {
		eprintf("Unknown hostname: %s\n", hostname);
		return -1;
	}
	memcpy(addr, ent->h_addr, sizeof(*addr));
	return 0;
}

int parse_hostsrv(const char * hostsrv, uint32_t * addr, int * port)
{
	char * addr_str, * port_str, * endptr;
	char * index;
	index = strchr(hostsrv, ':');
	if(index == NULL || index == hostsrv || *(index+1) == '\0') {
		eprintf(HERE "Invalid Arguments\n");
		return -1;
	}
	addr_str = strndup(hostsrv, index - hostsrv);
	port_str = ++index;
	if(resolve_host(addr_str, addr) != 0) {
		eprintf(HERE "Invalid Arguments\n");
		free(addr_str);
		return -1;
	}
	free(addr_str);
	*port = strtol(port_str, &endptr, 10);
	if(*endptr != '\0') {
		eprintf(HERE "Invalid Arguments %s\n", port_str);
		return -1;
	}
	return 0;
}

int parse_idset(char * str, vnode_idset * IDset)
{
	char * token, * ntok, * endptr;
	long int u,v;

	*IDset = 0;

	while ((token = strsep(&str, ",")) != NULL) {
		if(!isdigit(*token))
			return -1;
		if ((ntok = strchr(token, '-')) != NULL) {
			u = strtol(token, &endptr, 10);
			if (*endptr != '-')
				return -1;
			v = strtol(ntok + 1, &endptr, 10);
			if (*endptr != '\0')
				return -1;
			if(u > v || (0 > u || u > VNODE_ID_MAX)
				 || (0 > v || v > VNODE_ID_MAX))
				return -1;
			for (;u <= v; u++) {
				if (contain_IDset(u, *IDset))
					return -1;
				append_IDset(u, *IDset);
			}
		} else {
			u = strtol(token, &endptr, 10);
			if (*endptr != '\0')
				return -1;
			if(0 > u || u > VNODE_ID_MAX)
				return -1;
			if (contain_IDset(u, *IDset))
				return -1;
				append_IDset(u, *IDset);
		}
	}
	return 0;
}

int parse_nodespec(char * node, char **hb_node_name, vnode_idset * idset)
{
	const char delim[] = "/";
	char * sp;

	if(node == NULL)
		return -1;
	*hb_node_name = strtok_r(node, delim, &sp);
	if(strlen(*hb_node_name) == 0 || parse_idset(strtok_r(NULL, delim, &sp), idset) != 0) {
		return -1;
	}
	return 0;
}


