#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include "common.h"
#include "dtsadmin_result_handler.h"
#include "dtsd_fifocom.h"

/* file descriptor of fifo. */
static int rfd;
static int wfd;

static
int open_fifo()
{
	struct stat st;

	memset(&st, 0, sizeof(struct stat));

	if (stat(DTSD_FIFO_IN, &st) != 0 || S_ISFIFO(st.st_mode) != 1) {
		return -1;
	}

	if ((wfd = open(DTSD_FIFO_IN, O_WRONLY)) == -1) {
		perror("open");
		return -1;
	}

	if (stat(DTSD_FIFO_RES, &st) != 0 || S_ISFIFO(st.st_mode) != 1) {
		return -1;
	}

	if ((rfd = open(DTSD_FIFO_RES, O_RDONLY)) == -1) {
		perror("open");
		return -1;
	}


	return 0;
}

static
int close_fifo()
{
	struct stat st;

	memset(&st, 0, sizeof(struct stat));

	if (stat(DTSD_FIFO_IN, &st) != 0 && S_ISFIFO(st.st_mode) != 1) {
		return 0;
	}

	if (close(wfd) < 0) {
		return -1;
	}

	if (stat(DTSD_FIFO_RES, &st) != 0 && S_ISFIFO(st.st_mode) != 1) {
		return 0;
	}

	if (close(rfd) < 0) {
		return -1;
	}

	return 0;
}

static
int send_with_fifo(const char * str)
{
	int cwlen = 0, str_len = 0, twlen = 0;

	if (str == NULL || strnlen(str, MAX_OPT_LEN) == 0) {
		return -1;
	} else {
		str_len = strnlen(str, MAX_MSG_SIZE);
	}

	/* total wrote length. */
	twlen = 0;
	while (twlen < str_len) {
		/* write and get current wrote length. */
		cwlen = write(wfd, (str + twlen), strnlen(str + twlen, str_len - twlen));
		fsync(wfd);
		if (cwlen < 0) {
			return -1;
		}
		twlen += cwlen;
	}

	return 0;
}

static
int recv_with_fifo(char * str)
{
	int crlen = 0, trlen = 0;
	/* read from stdin and write to fifo. */
	for(;;) {
		crlen = read(rfd, str + trlen, 1);
		if(str[trlen] == EOF) {
			str[trlen] = '\0';
			break;
		}
		trlen++;
	}
	return 0;
}

int dtsadmin_init_fifocom(const char * progName, int debugMode)
{
	if (open_fifo() != 0) {
		return -1;
	} else {
		return 0;
	}
}

int dtsadmin_fin_fifocom(const char * progName, int debugMode)
{
	if (close_fifo() != 0) {
		return -1;
	} else {
		return 0;
	}
}

int dtsadmin_send_request_fifocom(const struct options * opt)
{
	char opt_str[MAX_OPT_LEN] = "";

	if (opt2str(opt, opt_str, INNER_OPT_PARSE) != 0) {
		return -1;
	}

	if (send_with_fifo(opt_str) != 0) {
		return -1;
	}
	return 0;
}

/* correct the processed result of all dtsd using fifo. */
int dtsadmin_receive_result_fifocom(char ** result_strv[], int * result_cnt)
{
	int ret = 0, i = 0, cnt = 0;
	char str[MAX_MSG_SIZE] = "";
	char ** res = NULL;
	char * tp = NULL;

	if (recv_with_fifo(str) != 0) {
		return -1;
	}

	res = (char **) malloc ((sizeof(char *)) * MAX_NUM_RESULTS);
	if (res == NULL) {
		return -1;
	}
	for (i = 0; i < MAX_NUM_RESULTS; i++) {
		res[i] = (char *) malloc((sizeof(char)) * MAX_CMD_RES_SIZE);
		if (res[i] == NULL) {
			ret = -1;
			goto free4failed;
		}
	}

	tp = strtok(str, TOKEN_RESULT);
	if (tp != NULL) {
		res[cnt] = strdup(tp);
		cnt++;
	} else {
		ret = -1;
		goto free4failed;
	}

	while (tp != NULL) {
		tp = strtok(NULL, TOKEN_RESULT);
		if (tp == NULL) {
			break;
		}

		strcpy(res[cnt], tp);
		cnt++;
	}

	*result_cnt = cnt;
	*result_strv = res;

	ret = 0;
	goto free4success;

free4success:
	for (i = cnt; i < MAX_NUM_RESULTS; i++) {
		free(res[i]);
		res[i] = NULL;
	}
	return ret;
free4failed:
	for (i = 0; i < MAX_NUM_RESULTS; i++) {
		free(res[i]);
		res[i] = NULL;
	}
	free(res);
	res = NULL;
	*result_strv = NULL;
	return ret;
}

