#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/select.h>
#include <unistd.h>
#include <errno.h>

#include "common.h"
#include "dtsd_common.h"
#include "dtsd_util.h"
#include "dtsd_optparser.h"
#include "dtsd_worker.h"
#include "dtsd_mediate.h"

#define DTSADMIN_ERROR_MESSAGE ">>\nHOST-host\nRESULT-NG\nERR-noerr\nMSG-nomsg\n"

static int read_line(int fd, char * dst, int size)
{
	char buf[1];
	int i = 0;
	for(;;) {
		if(read(fd, buf, 1) != 1) {
			return -1;
		}
		if(buf[0] == '\n') {
			dst[i] = '\0';
			return i;
		} else {
			dst[i] = buf[0];
		}
		i++;
		if(i == size) {
			dst[i + 1] = '\0';
			return size;
		}
	}
}

void remove_fifo()
{
	struct stat st;

	if (stat(DTSD_FIFO_RES, &st) == 0 || S_ISFIFO(st.st_mode) == 1) {
		if (unlink(DTSD_FIFO_RES) != 0) {
			/* continue for remove DTSD_FIFO_IN. */
			debug_fprintf(stderr, "dtsd: failed to remove " DTSD_FIFO_RES "\n");
		}
	}
	if (stat(DTSD_FIFO_IN, &st) == 0 || S_ISFIFO(st.st_mode) == 1) {
		if (unlink(DTSD_FIFO_IN) != 0) {
			debug_fprintf(stderr, "dtsd: failed to remove " DTSD_FIFO_IN"\n");
		}
	}
}

void * dtsd_mediate(void * args)
{
	/* This thread is a mediator between dtsdadmin and dtsd. */
	/* Wait dtsdadmin be executed, and receive its message. */
	// assert(args != NULL);

	int rfd, wfd, ret;
	struct dtsd_config * conf;
	const char end[2] = {EOF, '\0'};
	conf = (struct dtsd_config *) args;

	if(mkfifo(DTSD_FIFO_IN, S_IRWXU | S_IRWXG) != 0) {
		if(conf->force_flag == 1 && errno == EEXIST) {
			if(unlink(DTSD_FIFO_IN) != 0) {
				eprintf(HERE "failed to remove old FIFO.\n");
				exit(1);
			}
			debug_printf(HERE "removed old FIFO.\n");
		}
		if(mkfifo(DTSD_FIFO_IN, S_IRWXU | S_IRWXG) != 0) {
			perror("mkfifo");
			return NULL;
		}
	}
	if(mkfifo(DTSD_FIFO_RES, S_IRWXU | S_IRWXG) != 0) {
		if(conf->force_flag == 1 && errno == EEXIST) {
			if(unlink(DTSD_FIFO_RES) != 0) {
				eprintf(HERE "failed to remove old FIFO.\n");
				exit(1);
			}
			debug_printf(HERE "removed old FIFO.\n");
		}
		if(mkfifo(DTSD_FIFO_RES, S_IRWXU | S_IRWXG) != 0) {
			goto remove_fifo;
		}
	}

	for(;;) {
		fd_set rfds;
		char buffer[MAX_LEN_OPTV_VAL];
		char ** res;
		int resnum, i;
		struct timeval tv;
		struct options opt;

		if((rfd = open(DTSD_FIFO_IN, O_RDONLY)) == -1) {
			perror("open");
			exit(1);
		}
		if((wfd = open(DTSD_FIFO_RES, O_WRONLY)) == -1) {
			perror("open");
			exit(1);
		}

		tv.tv_sec = 3;
		tv.tv_usec = 0;

		FD_ZERO(&rfds);
		FD_SET(rfd, &rfds);

		ret = select(FD_SETSIZE, &rfds, NULL, NULL, &tv);

		if(ret == -1)
			perror("select()");
		if(ret) {
			if(read_line(rfd, buffer, sizeof(buffer)) < 0) {
				debug_printf(HERE "failed to read from " DTSD_FIFO_IN "\n");
				goto close_fd;
			}
			debug_printf(HERE "read from FIFO \"%s\"\n", buffer);
			// TODO FIX
			if(strlen(buffer) == 0)
				goto close_fd;
			if(str2opt(buffer, &opt, INNER_OPT_PARSE) != 0) {
				const char error[] = "Ignore unrecoginizable options.";
				eprintf("%s\n", error);
				if(write(wfd, error, strlen(error)) == -1) {
					perror("write");
				}
				goto close_fd;
			}
			debug_printf(HERE "succeded to convert command\n");
			if(dtsd_send_and_exec(&opt, &res, &resnum)) {
				const char error[] = "Failed in command execution.";
				eprintf("%s\n", error);
				if(write(wfd, error, strlen(error)) == -1) {
					perror("write");
				}
			}
			debug_printf(HERE "succeded to send command and receive %d results.\n", resnum);
			if(resnum == 0) {
				if(write(wfd, DTSADMIN_ERROR_MESSAGE, sizeof(DTSADMIN_ERROR_MESSAGE)) == -1) {
					perror("write");
					goto close_fd;
				}
			} else {
				for(i = 0; i < resnum; i++) {
					if(write(wfd, res[i], strlen(res[i])) == -1) {
						perror("write");
					}
					debug_printf(HERE "write result to FIFO \"%s\".\n", res[i]);
					free(res[i]);
				}
				free(res);
			}
			if(write(wfd, end, strlen(end)) == -1) {
				perror("write");
			}
			fsync(wfd);
		}
close_fd:
		close(rfd);
		close(wfd);
		// TODO FIX
	}
remove_fifo:
	remove_fifo();
	return NULL;
}
