#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "dtsadmin_codes.h"
#include "dtsd_common.h"
#include "dtsd_optparser.h"

/*
 * 1: must
 * x: illegal
 * +: optional
 * i: used only inner cmd.
 */
const char opt_cmb_tbl[NUM_CMD][NUM_OPT] =
{
/*                       -i   -M   -m   -f   -r   -s   -c   -n   -p   -v   --debug */
/* NONE             */ { 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', '+'},
/* ADD_CONF         */ { '1', '+', 'i', '1', '1', '+', 'x', 'x', 'i', 'x', '+'},
/* ADD_NODE         */ { 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', '+'},
/* DEL_CONF         */ { 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', '+'},
/* LIST             */ { 'x', 'x', 'x', 'x', 'x', 'x', '+', '+', 'x', '+', '+'},
/* CLEAR            */ { 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', '+'},
/* SAVE             */ { 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', '+'},
/* RESTORE          */ { 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', '+'},
/*                                inner cmd                                   */
/* GET_PROXY_PORT   */ { 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', '+'},
/* BIND_SOCK        */ { 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', '1', 'x', '+'},
/* UNBIND_SOCK      */ { 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', '1', 'x', '+'},
/* SET_MACADDR      */ { '1', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', '+'},
#ifdef DEBUG
/* PRINTCONFIG      */ { 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', '+'},
/* SHOWDATE         */ { 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', 'x', '+'},
#endif
};

/* for debug. */
void print_opt(const struct options * opt)
{
	printf("opt->cmd: %d\n", opt->cmd);
	printf("opt->is_inner_cmd: %d\n", opt->is_inner_cmd);
	printf("opt->host_srv: %s\n", opt->host_srv);
	printf("opt->iface: %s\n", opt->iface);
	printf("opt->netmask: %s\n", opt->netmask);
	printf("opt->macaddr: %s\n", opt->macaddr);
	printf("opt->fakemac: %d\n", opt->fakemac);
	printf("opt->ri: %s\n", opt->ri);
	printf("opt->ri_opt: %s\n", opt->ri_opt);
	printf("opt->nodes: %s\n", opt->nodes);
	printf("opt->proxy_port: %d\n", opt->proxy_port);
	printf("opt->chash: %d\n", opt->chash);
	printf("opt->dead_node: %d\n", opt->dead_node);
	printf("opt->debug: %d\n", opt->debug);
}

static
int validate_host_srv_val(const char * host_srv)
{
	uint32_t addr = 0;
	int port = 0;

	if (host_srv == NULL
		|| strnlen(host_srv, MAX_LEN_HOST_SRV) == 0) {
		return -1;
	}

	if (parse_hostsrv(host_srv, &addr, &port) != 0) {
		return -1;
	}

	return 0;
}

static
int validate_iface_val(const char * iface)
{
	if (iface == NULL
		|| strnlen(iface, MAX_LEN_IFACE) == 0) {
		return -1;
	}
	return 0;
}

static
int validate_netmask_val(const char * netmask)
{
	if (netmask == NULL
		|| strnlen(netmask, MAX_LEN_NETMASK) == 0) {
		return -1;
	}
	return 0;
}

static
int validate_macaddr_val(const char * macaddr)
{
	if (macaddr == NULL
		|| strnlen(macaddr, MAX_LEN_MACADDR) == 0) {
		return -1;
	}
	return 0;
}

static
int validate_fakemac_val(int fakemac)
{
	if (fakemac < MIN_VAL_FAKEMAC || fakemac > MAX_VAL_FAKEMAC) {
		return -1;
	}
	return 0;
}

static
int validate_ri_val(const char * ri)
{
	if (ri == NULL
		|| strnlen(ri, MAX_LEN_RI) == 0) {
		return -1;
	}
	return 0;
}

static
int validate_ri_opt_val(const char * ri_opt)
{
	return 0;
}

static
int validate_nodes_val(const char * nodes)
{
	char * tp = NULL;
	char * tmp_nodes = NULL;

	if (nodes == NULL
		|| strnlen(nodes, MAX_LEN_NODES) == 0) {
		return -1;
	}

	tmp_nodes = strdup(nodes);
	if (tmp_nodes == NULL) {
		return -1;
	}

	tp = strtok(tmp_nodes, TOKEN_COLON);
	if (tp == NULL) {
		return -1;
	}

	while((tp = strtok(NULL, TOKEN_COLON)) != NULL) {
		char * hb_node_name;
		vnode_idset idset;
		if(parse_nodespec(tp, &hb_node_name, &idset) != 0) {
			free(tmp_nodes);
			return -1;
		}
	}
	free(tmp_nodes);
	return 0;
}

static
int validate_proxy_port_val(int proxy_port)
{
	if (proxy_port < MIN_VAL_PROXY_PORT || proxy_port > MAX_VAL_PROXY_PORT) {
		return -1;
	}
	return 0;
}

static
int validate_chash_val(int chash)
{
	if (chash < MIN_VAL_CHASH || chash > MAX_VAL_CHASH) {
		return -1;
	}
	return 0;
}

static
int validate_dead_node_val(int dead_node)
{
	if (dead_node < MIN_VAL_DEAD_NODE || dead_node > MAX_VAL_DEAD_NODE) {
		return -1;
	}
	return 0;
}

/* validate option's combination. */
static
int validate_opt_cmb(int opt_cmb, const struct options * opt, int parse_method)
{
	int i = 0;

	if (opt == NULL) {
		return -1;
	}

	for (i = 0; i < NUM_OPT; i++) {
		if ((opt_cmb & (1 << i))) {
			/* 1: must option check. */
			if (opt_cmb_tbl[opt->cmd][i] == '1') {
				// ok. nothing to do.
			/* +: optional option check. */
			} else if (opt_cmb_tbl[opt->cmd][i] == '+') {
				// ok. nothing to do.
			/* x: illegal option check. */
			} else if (opt_cmb_tbl[opt->cmd][i] == 'x') {
				// invalid option specified.
				return -1;
			/* inner option check. */
			} else if (opt_cmb_tbl[opt->cmd][i] == 'i') {
				if (parse_method == INNER_OPT_PARSE) {
					// ok. nothing to do.
				} else {
					return -1;
				}
			}
		} else {
			/* 1: must option check */
			if (opt_cmb_tbl[opt->cmd][i] == '1') {
				return -1;
			/* +: optional option check. */
			} else if (opt_cmb_tbl[opt->cmd][i] == '+') {
				// ok. nothing to do.
			/* x: illegal option check. */
			} else if (opt_cmb_tbl[opt->cmd][i] == 'x') {
				// ok. nothing to do.
			/* inner option check. */
			} else if (opt_cmb_tbl[opt->cmd][i] == 'i') {
				if (parse_method == INNER_OPT_PARSE) {
					return -1;
				} else {
					// ok. nothing to do.
				}
			}
		}
	}
	return 0;
}

/* validate option's value. */
static
int validate_opt_val(const struct options * opt, int parse_method)
{
	if (opt == NULL) {
		return -1;
	}

	switch(opt->cmd) {
		case ADD_CONF:
			if (validate_host_srv_val(opt->host_srv) != 0 ||
				validate_iface_val(opt->iface) != 0 ||
				/* netmask is omitable, but value is set by set_default_val(). */
				validate_netmask_val(opt->netmask) != 0 ||
				validate_fakemac_val(opt->fakemac) != 0 ||
				validate_ri_opt_val(opt->ri_opt) != 0) {
				return -1;
			}
			/* if specified the ri_opt, then check it. */
			if (strnlen(opt->ri_opt, MAX_LEN_RI_OPT) != 0) {
				if (validate_ri_val(opt->ri) != 0) {
					return -1;
				}
			}
			/* if specified the nodes, then check it. */
			if (strnlen(opt->nodes, MAX_LEN_NODES) != 0) {
				if (validate_nodes_val(opt->nodes) != 0) {
					return -1;
				}
			}
			if (parse_method == INNER_OPT_PARSE) {
				if (validate_macaddr_val(opt->macaddr) != 0 ||
					validate_proxy_port_val(opt->proxy_port) != 0) {
					return -1;
				}
			}
			break;
		case ADD_NODE:
		case DEL_NODE:
			if (validate_nodes_val(opt->nodes) != 0) {
				return -1;
			}
			if (validate_host_srv_val(opt->host_srv) != 0) {
				return -1;
			}
			break;
		case LIST:
			if (validate_dead_node_val(opt->dead_node) != 0) {
				return -1;
			}
			if (validate_chash_val(opt->chash) != 0) {
				return -1;
			}
			break;
		case BIND_SOCK:
		case UNBIND_SOCK:
			if (parse_method == INNER_OPT_PARSE) {
				if (validate_proxy_port_val(opt->proxy_port) != 0) {
					return -1;
				}
			}
			break;
		case RESTORE:
		case CLEAR:
		case SAVE:
			break;
		case SET_MACADDR:
			if (validate_iface_val(opt->iface) != 0) {
				return -1;
			}
		case NONE:
		default:
			break;
	}
	return 0;
}

int set_default_val(struct options * opt) {
	if (opt == NULL) {
		return -1;
	}

	switch(opt->cmd) {
		case ADD_CONF:
			if (opt->netmask == NULL
				|| strnlen(opt->netmask, MAX_LEN_NETMASK) == 0) {
				strcpy(opt->netmask, "255.255.255.0");
			}
			break;
		case ADD_NODE:
		case DEL_NODE:
		case LIST:
		case BIND_SOCK:
		case UNBIND_SOCK:
		case RESTORE:
		case CLEAR:
		case SAVE:
		case SET_MACADDR:
		case NONE:
		default:
			break;
	}
	return 0;
}

/* convert option vector to struct options.              */
/* parse_method is following.                            */
/*                                                       */
/* OUTER_OPT_PARSE: parse for dtsadmin's run argument.   */
/* INNER_OPT_PARSE: parse for inner argument for system. */
int optv2opt(char * optv[], struct options * opt, int parse_method)
{
	int opt_cmb = 0;

	if (optv == NULL || opt == NULL) {
		return -1;
	}

	memset(opt, 0, sizeof(struct options));

	// get command type.
	if (strncmp(OPT_ADD_CONF, *optv, strlen(OPT_ADD_CONF)) == 0) {
		opt->cmd = ADD_CONF;
		if (*(optv + 1) != NULL) {
			snprintf(opt->host_srv, MAX_LEN_HOST_SRV,
						 "%s", *(optv + 1));
			optv++;
		}
	} else if (strncmp(OPT_ADD_NODE, *optv, strlen(OPT_ADD_NODE)) == 0) {
		opt->cmd = ADD_NODE;
		if (*(optv + 1) != NULL) {
			snprintf(opt->host_srv, MAX_LEN_HOST_SRV,
						"%s", *(optv + 1));
			optv++;
		}
	} else if (strncmp(OPT_DEL_NODE, *optv, strlen(OPT_DEL_NODE)) == 0) {
		opt->cmd = DEL_NODE;
		if (*(optv + 1) != NULL) {
			snprintf(opt->host_srv, MAX_LEN_HOST_SRV,
						"%s", *(optv + 1));
			optv++;
		}
	} else if (strncmp(OPT_LIST, *optv, strlen(OPT_LIST)) == 0) {
		opt->cmd = LIST;
	} else if (strncmp(OPT_CLEAR, *optv, strlen(OPT_CLEAR)) == 0) {
		opt->cmd = CLEAR;
	} else if (strncmp(OPT_SAVE, *optv, strlen(OPT_SAVE)) == 0) {
		opt->cmd = SAVE;
	} else if (strncmp(OPT_RESTORE, *optv, strlen(OPT_RESTORE)) == 0) {
		opt->cmd = RESTORE;
	} else if (strncmp(OPT_GET_PROXY_PORT, *optv, strlen(OPT_GET_PROXY_PORT)) == 0) {
		opt->cmd = GET_PROXY_PORT;
		opt->is_inner_cmd = INNER_CMD;
	} else if (strncmp(OPT_BIND_SOCK, *optv, strlen(OPT_BIND_SOCK)) == 0) {
		opt->cmd = BIND_SOCK;
		opt->is_inner_cmd = INNER_CMD;
	} else if (strncmp(OPT_UNBIND_SOCK, *optv, strlen(OPT_UNBIND_SOCK)) == 0) {
		opt->cmd = UNBIND_SOCK;
		opt->is_inner_cmd = INNER_CMD;
	} else if (strncmp(OPT_SET_MACADDR, *optv, strlen(OPT_SET_MACADDR)) == 0) {
		opt->cmd = SET_MACADDR;
		opt->is_inner_cmd = INNER_CMD;
#ifdef DEBUG
	} else if (strncmp(OPT_PRINT_CONFIG, *optv, strlen(OPT_PRINT_CONFIG)) == 0) {
		opt->cmd = PRINTCONF;
	} else if (strncmp(OPT_SHOW_DATE, *optv, strlen(OPT_SHOW_DATE)) == 0) {
		opt->cmd = SHOWDATE;
#endif
	} else {
		return -1;
	}

	optv++;

	// get sub options.
	while(*optv != NULL) {
		if (strcmp(OPT_IFACE, *optv) == 0) {
			opt_cmb = opt_cmb | 1 << SHIFT_BIT_IFACE;
			if (*(optv + 1) != NULL) {
				snprintf(opt->iface, MAX_LEN_IFACE,
						"%s", *(optv + 1));
				optv++;
			}
		} else if (strcmp(OPT_NETMASK, *optv) == 0) {
			opt_cmb = opt_cmb | 1 << SHIFT_BIT_NETMASK;
			if (*(optv + 1) != NULL) {
				snprintf(opt->netmask, MAX_LEN_NETMASK,
						"%s", *(optv + 1));
				optv++;
			}
		} else if (strcmp(OPT_MACADDR, *optv) == 0) {
			opt_cmb = opt_cmb | 1 << SHIFT_BIT_MACADDR;
			if (*(optv + 1) != NULL) {
				snprintf(opt->macaddr, MAX_LEN_MACADDR,
						"%s", *(optv + 1));
				optv++;
			}
		} else if (strcmp(OPT_FAKEMAC, *optv) == 0) {
			opt_cmb = opt_cmb | 1 << SHIFT_BIT_FAKEMAC;
			if (*(optv + 1) != NULL) {
				if (strcmp("on", *(optv + 1)) == 0) {
					opt->fakemac = 1;
				} else {
					opt->fakemac = 0;
				}
				optv++;
			}
		} else if (strcmp(OPT_RI, *optv) == 0) {
			opt_cmb = opt_cmb | 1 << SHIFT_BIT_RI;
			if (*(optv + 1) != NULL) {
				snprintf(opt->ri, MAX_LEN_RI,
						"%s", *(optv + 1));
				optv++;
			}
		} else if (strcmp(OPT_RI_OPT, *optv) == 0) {
			opt_cmb = opt_cmb | 1 << SHIFT_BIT_RI_OPT;
			if (*(optv + 1) != NULL) {
				snprintf(opt->ri_opt, MAX_LEN_RI_OPT,
						"%s", *(optv + 1));
				optv++;
			}
		} else if (strcmp(OPT_CHASH, *optv) == 0) {
			opt_cmb = opt_cmb | 1 << SHIFT_BIT_CHASH;
			opt->chash = 1;
		} else if (strcmp(OPT_DEAD_NODE, *optv) == 0) {
			opt_cmb = opt_cmb | 1 << SHIFT_BIT_DEAD_NODE;
			opt->dead_node = 1;
		} else if (strcmp(OPT_PROXY_PORT, *optv) == 0) {
			opt_cmb = opt_cmb | 1 << SHIFT_BIT_PROXY_PORT;
			if (*(optv + 1) != NULL) {
				opt->proxy_port = atoi(*(optv + 1));
				optv++;
			}
		} else if (strcmp(OPT_VERBOSE, *optv) == 0) {
			opt_cmb = opt_cmb | 1 << SHIFT_BIT_VERBOSE;
			opt->verbose = 1;
		} else if (strcmp(OPT_DEBUG, *optv) == 0) {
			opt->debug = 1;
		} else {
			if (opt->cmd == ADD_CONF || opt->cmd == ADD_NODE ||
				opt->cmd == DEL_NODE) {
				snprintf(opt->nodes, MAX_LEN_NODES,
						"%s", *optv);
				break;
			}
			return -1;
		}
		optv++;
	}

	if (set_default_val(opt) != 0) {
		return -1;
	}

	if (validate_opt_cmb(opt_cmb, opt, parse_method) != 0) {
		return -1;
	} else if (validate_opt_val(opt, parse_method) != 0) {
		return -1;
	}
	return 0;
}

/* convert struct options to string.                      */
/* OUTER_OPT_PARSE: output the dtsadmin's run argument.   */
/* INNER_OPT_PARSE: output the inner argument for system. */
int opt2str(const struct options * opt, char * opt_str, int parse_method)
{
	char opt_fakemac[MAX_LEN_FAKEMAC] = "";
	char opt_proxy_port[MAX_LEN_PROXY_PORT] = "";
	char opt_macaddr[MAX_LEN_MACADDR] = "";
	char opt_ri_opt[MAX_LEN_RI_OPT] = "";
	char opt_chash[MAX_LEN_CHASH] = "";
	char opt_dead_node[MAX_LEN_DEAD_NODE] = "";

	if (opt == NULL || opt_str == NULL) {
		return -1;
	}

	switch(opt->cmd) {
		case ADD_CONF:
			if (opt->fakemac) {
				snprintf(opt_fakemac, MAX_LEN_FAKEMAC,
						"%s on", OPT_FAKEMAC);
			} else {
				snprintf(opt_fakemac, MAX_LEN_FAKEMAC,
						"%s off", OPT_FAKEMAC);
			}
			if (opt->ri_opt != NULL) {
				snprintf(opt_ri_opt, MAX_LEN_RI_OPT,
						"%s \"%s\"", OPT_RI_OPT, opt->ri_opt);
			}

			if (parse_method == INNER_OPT_PARSE) {
				snprintf(opt_macaddr, MAX_LEN_MACADDR, "%s %s",
						OPT_MACADDR, opt->macaddr);
			}

			if (parse_method == INNER_OPT_PARSE) {
				snprintf(opt_proxy_port, MAX_LEN_PROXY_PORT, "%s %d",
						OPT_PROXY_PORT, opt->proxy_port);
			}

			snprintf(opt_str, MAX_OPT_LEN, "%s %s %s %s %s %s %s %s %s %s %s %s %s\n",
						OPT_ADD_CONF,
						opt->host_srv,
						OPT_IFACE,
						opt->iface,
						OPT_NETMASK,
						opt->netmask,
						opt_macaddr,
						opt_proxy_port,
						opt_fakemac,
						OPT_RI,
						opt->ri,
						opt_ri_opt,
						opt->nodes);
			break;
		case ADD_NODE:
			snprintf(opt_str, MAX_OPT_LEN, "%s \"%s\" \"%s\"\n",
						OPT_ADD_NODE,
						opt->host_srv,
						opt->nodes);
			break;
		case DEL_NODE:
			snprintf(opt_str, MAX_OPT_LEN, "%s \"%s\" \"%s\"\n",
						OPT_DEL_NODE,
						opt->host_srv,
						opt->nodes);
			break;
		case LIST:
			if (opt->chash) {
				snprintf(opt_chash, MAX_LEN_CHASH,
						"%s", OPT_CHASH);
			}
			if (opt->dead_node) {
				snprintf(opt_dead_node, MAX_LEN_DEAD_NODE,
						"%s", OPT_DEAD_NODE);
			}
			snprintf(opt_str, MAX_OPT_LEN, "%s %s %s\n",
						OPT_LIST,
						opt_chash,
						opt_dead_node);
			break;
		case CLEAR:
			snprintf(opt_str, MAX_OPT_LEN, "%s\n", OPT_CLEAR);
			break;
		case SAVE:
			snprintf(opt_str, MAX_OPT_LEN, "%s\n", OPT_SAVE);
			break;
		case RESTORE:
			snprintf(opt_str, MAX_OPT_LEN, "%s\n", OPT_RESTORE);
			break;
		case GET_PROXY_PORT:
			snprintf(opt_str, MAX_OPT_LEN, "%s\n",
						OPT_GET_PROXY_PORT);
			break;
		case BIND_SOCK:
			snprintf(opt_str, MAX_OPT_LEN, "%s %s %d\n",
						OPT_BIND_SOCK, OPT_PROXY_PORT, opt->proxy_port);
			break;
		case UNBIND_SOCK:
			snprintf(opt_str, MAX_OPT_LEN, "%s %s %d\n",
						OPT_UNBIND_SOCK, OPT_PROXY_PORT, opt->proxy_port);
			break;
		case SET_MACADDR:
			snprintf(opt_str, MAX_OPT_LEN, "%s %s %s\n", OPT_SET_MACADDR, OPT_IFACE, opt->iface);
			break;
#ifdef DEBUG
		case PRINTCONF:
			snprintf(opt_str, MAX_OPT_LEN, "%s\n", OPT_PRINT_CONFIG);
			break;
		case SHOWDATE:
			snprintf(opt_str, MAX_OPT_LEN, "%s\n", OPT_SHOW_DATE);
			break;
#endif
		case NONE:
		default:
			break;
	}
	return 0;
}

/* convert string to struct options.                     */
/* OUTER_OPT_PARSE: parse for dtsadmin's run argument.   */
/* INNER_OPT_PARSE: parse for inner argument for system. */
int str2opt(const char * str, struct options * opt, int parse_method)
{
	int i = 0, optc = 0, str_len = 0;
	int start_idx = 0, st_flag = 0, d_flag = 0;
	char optv[MAX_LEN_OPTV][MAX_LEN_OPTV_VAL];
	const char *optvp[MAX_LEN_OPTV];

	if (str == NULL || opt == NULL) {
		return -1;
	}

	str_len = strlen(str);
	optc = 0;
	for (i = 0; i < str_len; i++) {
		if (st_flag == 0) {
			// find not space
			if (str[i] != TOKEN_SPACE) {
				st_flag = 1;
				start_idx = i;

				if (str[i] == TOKEN_DQUOTE) {
					d_flag = 1;
					// ignore dquote
					start_idx++;
				}
			}
		} else {
			if (d_flag == 0) {
				if (str[i] == TOKEN_SPACE) {
					// from start_idx to i copy
					st_flag = 0;
					strncpy(optv[optc], str + start_idx, i - start_idx);
					optv[optc][i - start_idx] = '\0';
					if (optv[optc][strlen(optv[optc]) - 1] == '\n') {
						optv[optc][strlen(optv[optc]) - 1] = '\0';
					}
					optc++;
					continue;
				}
			} else {
				if (str[i] == TOKEN_DQUOTE) {
					// from start_idx to i copy
					d_flag = 0;
					st_flag = 0;
					strncpy(optv[optc], str + start_idx, i - start_idx);
					optv[optc][i - start_idx] = '\0';
					if (optv[optc][strlen(optv[optc]) - 1] == '\n') {
						optv[optc][strlen(optv[optc]) - 1] = '\0';
					}
					optc++;
					continue;
				}
			}

			if (i == (str_len - 1)) {
				/* not complete double quote. */
				if (d_flag == 1) {
					return -1;
				}
				// from start_idx to (i + 1) copy
				strncpy(optv[optc], str + start_idx, i - start_idx + 1);
				optv[optc][i - start_idx + 1] = '\0';
				if (optv[optc][strlen(optv[optc]) - 1] == '\n') {
					optv[optc][strlen(optv[optc]) - 1] = '\0';
				}
				optc++;
			}
		}
	}

	for (i = 0; i < MAX_LEN_OPTV; i++) {
		optvp[i] = optv[i];
		if (i >= optc) {
			optvp[i] = NULL;
		}
	}

	return optv2opt(optvp, opt, parse_method);
}

/* the test driver. */
/*
int main(int argc, char * argv[])
{
	struct options opt
		= { 0, 0, "", "", "", "", 0, "", "", "", 0, 0, 0 };
//	char opt_str[] = "-A 192.168.1.1:80 -i eth0 -M 255.255.255.0 -f on -r httpri -s \"rr aaaa\" 10.0.0.1:10.0.0.2";
//	char opt_str[] = "-A 192.168.1.1:80 -i eth0 -f on -r httpri -s \"rr aaaa\" 10.0.0.1:10.0.0.2";
	char opt_str[] = "-A 192.168.1.1:80 -i eth0 -f on -r httpri 10.0.0.1:10.0.0.2";
//	char opt_str[] = "-A 192.168.1.1:80 -i eth0 -M 255.255.255.0 -m aa:bb:cc:dd:ee:ff -f on -r httpri -s \"rr aaaa\" -p 65532 10.0.0.1:10.0.0.2";
	char str[MAX_OPT_LEN];

	if (str2opt(opt_str, &opt, OUTER_OPT_PARSE) != 0) {
		printf("failed to str2opt.\n");
	}

	print_opt(&opt);

	if (opt2str(&opt, str, OUTER_OPT_PARSE) != 0) {
		printf("failed to opt2str.\n");
	}

	printf("%s\n", str);

	if (str2opt(str, &opt, OUTER_OPT_PARSE) != 0) {
		printf("failed to str2opt.\n");
	}

	print_opt(&opt);

	return 0;
}
*/

