#ifndef _DTSD_OPTPARSER_H
#define _DTSD_OPTPARSER_H

#include "common.h"

#ifdef DEBUG
# define NUM_CMD                  13
#else
# define NUM_CMD                  12
#endif /* DEBUG */
#define NUM_OPT                   11

#define OPT_ADD_CONF            "-A"
#define OPT_ADD_NODE            "-a"
#define OPT_DEL_NODE            "-d"
#define OPT_LIST                "-L"
#define OPT_CLEAR               "-C"
#define OPT_SAVE                "-S"
#define OPT_RESTORE             "-R"
#define OPT_GET_PROXY_PORT      "-G"
#define OPT_BIND_SOCK           "-B"
#define OPT_UNBIND_SOCK         "-U"
#define OPT_SET_MACADDR         "-H"
#define OPT_IFACE               "-i"
#define OPT_NETMASK             "-M"
#define OPT_MACADDR             "-m"
#define OPT_FAKEMAC             "-f"
#define OPT_RI                  "-r"
#define OPT_RI_OPT              "-s"
#define OPT_VERBOSE             "-v"
#define OPT_CHASH               "-c"
#define OPT_DEAD_NODE           "-n"
#define OPT_PROXY_PORT          "-p"
#define OPT_DEBUG               "--debug"
#ifdef DEBUG
#define OPT_PRINT_CONFIG        "--printf-config"
#define OPT_SHOW_DATE           "--show-date"
#endif

#define SHIFT_BIT_IFACE            0
#define SHIFT_BIT_NETMASK          1
#define SHIFT_BIT_MACADDR          2
#define SHIFT_BIT_FAKEMAC          3
#define SHIFT_BIT_RI               4
#define SHIFT_BIT_RI_OPT           5
#define SHIFT_BIT_CHASH            6
#define SHIFT_BIT_DEAD_NODE        7
#define SHIFT_BIT_PROXY_PORT       8
#define SHIFT_BIT_VERBOSE          9

#define INNER_CMD                  1

#define TOKEN_SPACE              ' '
#define TOKEN_DQUOTE             '"'

#define INNER_OPT_PARSE            1
#define OUTER_OPT_PARSE            0

enum cmd_type
{
	NONE           = 0,
	ADD_CONF       = 1,
	ADD_NODE       = 2,
	DEL_NODE       = 3,
	LIST           = 4,
	CLEAR          = 5,
	SAVE           = 6,
	RESTORE        = 7,
	GET_PROXY_PORT = 8,
	BIND_SOCK      = 9,
	UNBIND_SOCK    = 10,
	SET_MACADDR    = 11,
#ifdef DEBUG
	PRINTCONF,
	SHOWDATE,
#endif
};

struct options
{
	/* type of cmd. */
	enum cmd_type cmd;
	int is_inner_cmd;

	/* parameters. */
	char host_srv[MAX_LEN_HOST_SRV];
	char iface[MAX_LEN_IFACE];
	char netmask[MAX_LEN_NETMASK];
	char macaddr[MAX_LEN_MACADDR];
	int fakemac;
	char ri[MAX_LEN_RI];
	char ri_opt[MAX_LEN_RI_OPT];
	char nodes[MAX_LEN_NODES];
	int proxy_port;
	int chash;
	int dead_node;
	int verbose;
	int debug;
};

extern int optv2opt(char * argv[], struct options * opt, int parse_method);
extern int opt2str(const struct options * opt, char * str, int parse_method);
extern int str2opt(const char * str, struct options * opt, int parse_method);
extern void print_opt(const struct options * opt);

#endif

