#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <assert.h>
#include <dirent.h>
#include <dlfcn.h>
#include <stdint.h>
#include <stdbool.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <net/if.h>

#include "common.h"
#include "dtsd_common.h"
#include "dtsd_util.h"
#include "dtsd_ri_plugins.h"
#include "dtsd_plugin_loader.h"

struct ri_intf_container {
	struct ri_intf_container * next;
	void * handle;
	struct ri_intf * intf;
};

#define append_ri_intf_set(ri_intf_container) \
	ri_intf_set_last->next = (ri_intf_container);	\
	ri_intf_set_last = (ri_intf_container);\
	ri_intf_set_last->next = NULL

static struct ri_intf_container RI_INTF_SET_FIRST = {.next = NULL, .intf = NULL};
static struct ri_intf_container * ri_intf_set_last = & RI_INTF_SET_FIRST;

union { dts_dll_init_function func; void * obj; } alias;

int load_plugin(const char * dir, const char * filename)
{
	char * filepath;
	void * handle;
	char * error;
	// dts_dll_init_function create_ri_intf;
	struct ri_intf * (* create_ri_intf)(void);
	struct ri_intf * ri_intf;
	struct ri_intf_container * ri_intf_container;

	filepath = (char *) malloc (sizeof(char) * (strlen(dir) + strlen(filename) + 2));
	sprintf(filepath, "%s/%s", dir, filename);
	debug_fprintf(stderr, HERE "loading %s\n", filepath);

	/* if((handle = dlopen(filepath, RTLD_LAZY | RTLD_NODELETE)) == NULL) { */
	if((handle = dlopen(filepath, RTLD_NOW | RTLD_NODELETE)) == NULL) {
		debug_fprintf(stderr, HERE "%s\n", dlerror());
		return 1;
	}

	dlerror();

	alias.obj = dlsym(handle, "create_ri_intf");
	if ((error = dlerror()) != NULL)  {
		fprintf(stderr, HERE "%s\n", error);
		dlclose(handle);
		return 1;
	}
	if (alias.func == NULL) {
		fprintf(stderr, HERE "alias.func is NULL.\n");
		dlclose(handle);
		return 1;
	}
	create_ri_intf = alias.func;
	ri_intf = (*create_ri_intf)();

	ri_intf_container = (struct ri_intf_container *) malloc (sizeof(struct ri_intf_container));
	ri_intf_container->intf = ri_intf;
	ri_intf_container->handle = handle;
	append_ri_intf_set(ri_intf_container);

	debug_fprintf(stderr, HERE "load plugin %s\n", ri_intf->name);
	return 0;
}

// return the num plugin loaded.
int load_plugins()
{
	struct dirent **namelist;
	int n, l;

	// num plugin loaded.
	l = 0;
	if((n = scandir(DTS_PLUGIN_DIR, &namelist, 0, alphasort)) < 0) {
		perror("scandir");
	} else {
		while (n--) {
			if(*(namelist[n]->d_name) == '.') {
				free(namelist[n]);
				continue;
			}
			if (load_plugin(DTS_PLUGIN_DIR, namelist[n]->d_name) == 0) {
				l++;
			}
			free(namelist[n]);
		}
		free(namelist);
	}
	debug_printf(HERE "ri plugin loaded. num: %d\n", l);
	return l;
}

struct ri_intf_container * next_ri_intf_container (struct ri_intf_container * cnt)
{
	if (cnt == NULL) {
		return NULL;
	}
	return cnt->next;
}

struct ri_intf_container * first_ri_intf_container ()
{
	return RI_INTF_SET_FIRST.next;
}

struct ri_intf * get_ri_intf_container (struct ri_intf_container * cnt)
{
	if (cnt == NULL) {
		return NULL;
	}
	return cnt->intf;
}

struct ri_intf * find_ri_intf (const char * riname)
{
	struct ri_intf_container * i;
	struct ri_intf * ri;

	if(riname == NULL) {
		eprintf(HERE "Invalid Arugments.\n");
	}

	for(i = first_ri_intf_container();
		i != NULL;
		i = next_ri_intf_container(i)) {
		ri = get_ri_intf_container(i);
		if(strcmp(riname, ri->name) == 0) {
			debug_printf(HERE "find Request Inspector named \"%s\"\n", ri->name);
			return ri;
		}
	}
	eprintf(HERE "No RequestInspector named %s\n", riname);
	return NULL;
}
