#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <stdlib.h>
#include <ctype.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <assert.h>

#include "common.h"
#include "dtsd_common.h"
#include "dtsd_util.h"
#include "dtsd_optparser.h"
#include "dtsd_server_conf.h"
#include "dtsd_command_exec.h"
#include "dtsd_worker.h"

int NodeInfo2opt(const struct dtsd_node_info * info,
						struct options * opt)
{
	uint32_t addr = info->srv_conf->server_address;
	struct in_addr in_addr = {.s_addr = addr};
	const char * addr_str = inet_ntoa(in_addr);
	char vnode_str[MAX_LEN_NODES] = "";

	if(info == NULL) {
		eprintf(HERE "node infomation is not specified.\n");
	}
	if(opt == NULL) {
		eprintf(HERE "opt is not specified.\n");
	}
	serialize_idset(info->idset, vnode_str);
	opt->cmd = ADD_NODE;
	snprintf(opt->host_srv, sizeof(opt->host_srv), "%s:%d", addr_str, info->srv_conf->server_port);
	if (strlen(vnode_str) > 0) {
		snprintf(opt->nodes, sizeof(opt->nodes), "%s/%s", info->hb_node_name, vnode_str);
	} else {
		snprintf(opt->nodes, sizeof(opt->nodes), "%s", info->hb_node_name);
	}
	return 0;
}

int opt2NodeInfo(struct options * opt, struct dtsd_node_info * info)
{
	const char delim = '/';
	vnode_idset idset = 0;
	char * hbname, * idset_str, * nodes, * index;
	if(opt->cmd != ADD_NODE && opt->cmd != DEL_NODE) {
		eprintf(HERE "Invalid Arguments.\n");
		return -1;
	}
	nodes = opt->nodes;
	index = strchr(nodes, delim);
	if(index == NULL) {
		hbname = strdup(nodes);
		idset = 0;
	} else {
		hbname = strndup(nodes, index - nodes);
		idset_str = ++index;
		if(parse_idset(idset_str, &idset) != 0) {
			eprintf(HERE "Invalid Node Spec.\n");
			free(hbname);
			return -1;
		}
	}
	info->hb_node_name = hbname;
	info->idset = idset;
	info->next = NULL;
	info->srv_conf = NULL;
	info->hbc_node = find_or_add_hbc_node(hbname);
	return 0;
}

int serialize_NodeInfo(const struct dtsd_node_info * info, char * str)
{
	struct options opt;
	if(info == NULL || str == NULL) {
		eprintf(HERE "Invalid Arguments\n");
		return -1;
	}
	if(NodeInfo2opt(info, &opt) != 0) {
		eprintf(HERE "Invalid Arguments\n");
		return -1;
	}
	if(opt2str(&opt, str, INNER_OPT_PARSE) != 0) {
		eprintf(HERE "Invalid Arguments\n");
		return -1;
	}
	return 0;
}

int deserialize_NodeInfo(char * str, struct dtsd_node_info * info)
{
	struct options opt;
	if(info == NULL || str == NULL) {
		eprintf(HERE "Invalid Arguments\n");
		return -1;
	}
	if(str2opt(str, &opt, INNER_OPT_PARSE) != 0) {
		eprintf(HERE "Invalid Arguments\n");
		return -1;
	}
	if(opt2NodeInfo(&opt, info) != 0) {
		eprintf(HERE "Invalid Arguments\n");
		return -1;
	}
	return 0;
}

int parse_nodespec_list(char * node_list, struct dtsd_node_info ** head)
{
	char * token, * sp;
	struct dtsd_node_info * root = NULL, *i = NULL, *j;
	while((token = strtok_r(node_list, ":", &sp)) != NULL) {
		node_list = NULL;
		j = (struct dtsd_node_info *) malloc (sizeof(struct dtsd_node_info));
		memset(j, 0, sizeof(struct dtsd_node_info));
		char * hb_node_name;
		vnode_idset idset;
		if(parse_nodespec(token, &hb_node_name, &idset) != 0) {
			free_node_info(j);
			free(node_list);
			goto free_node_list;
		}
		j->hb_node_name = strdup(hb_node_name);
		j->idset = idset;
		if(root == NULL) {
			root = j;
			i = j;
		} else {
			i->next = j;
			i = j;
		}
	}
	*head = root;
	return 0;

free_node_list:
	if(root != NULL)
		free_node_info_list(root);
	return -1;
}

int free_node_info(struct dtsd_node_info * info)
{
	if(info == NULL) {
		return 0;
	}
	if(info->hb_node_name != NULL)
		free(info->hb_node_name);
	free(info);
	return 0;
}

int free_node_info_list(struct dtsd_node_info * info)
{
	struct dtsd_node_info * i, *j = NULL;
	i = info;
	for(i = info; i != NULL; i = i->next) {
		if(j != NULL)
			free_node_info(j);
		j = i;
	}
	return 0;
}

int ServerConf2opt(const struct server_conf * conf, struct options * opt)
{
	const char * addr_str, * netmask_str;
	struct in_addr in_addr;

	opt->cmd = ADD_CONF;
	in_addr.s_addr = conf->server_address;
	addr_str = inet_ntoa(in_addr);
	sprintf(opt->host_srv, "%s:%d", addr_str, conf->server_port);
	in_addr.s_addr = conf->netmask;
	netmask_str = inet_ntoa(in_addr);
	strcpy(opt->iface, conf->iface);
	strcpy(opt->netmask, netmask_str);
	opt->fakemac = conf->fakemac;
	strcpy(opt->ri, conf->riname);
	strcpy(opt->ri_opt, conf->ri_opt);
	strcpy(opt->macaddr, conf->hwaddr);
	opt->proxy_port = conf->proxy_port;
	opt->nodes[0] = '\0'; // don't use initiali node set
	return 0;
}

int opt2ServerConf(struct options * opt, struct server_conf * conf)
{
	char * hostsrv;
	struct dtsd_node_info * tmp_info, * i;
	hostsrv = opt->host_srv;
	if(parse_hostsrv(hostsrv, &(conf->server_address), &(conf->server_port)) != 0) {
		eprintf(HERE "Invalid Arguments\n");
		return -1;
	}
	if(resolve_host(opt->netmask, &(conf->netmask)) != 0) {
		eprintf(HERE "Invalid Arguments\n");
		return -1;
	}
	if(strlen(opt->macaddr) > 0)
		conf->hwaddr = strdup(opt->macaddr);
	if(strlen(opt->iface) <= 0) {
		eprintf(HERE "Invalide Arguments\n");
		return -1;
	}
	conf->iface = strdup(opt->iface);
	conf->fakemac = opt->fakemac;
	if(strlen(opt->ri) <= 0) {
		eprintf(HERE "Invalide Arguments\n");
		return -1;
	}
	conf->riname = strdup(opt->ri);
	conf->ri_opt = strdup(opt->ri_opt);
	conf->proxy_port = opt->proxy_port;
	conf->ri = NULL;
	conf->ri_intf = NULL;
	conf->next = NULL;
	conf->ni = NULL;

	tmp_info = (struct dtsd_node_info *) malloc (sizeof(struct dtsd_node_info));
	if (tmp_info == NULL) {
		eprintf(HERE "malloc failed.\n");
		return -1;
	}
	if (parse_nodespec_list(opt->nodes, &tmp_info) != 0) {
		eprintf(HERE "parse nodespec failed.\n");
		free(tmp_info);
		return -1;
	}

	add_NodeInfo(conf, tmp_info);

	for_each_node_info(i, conf) {
		/* hb_node_name is NULL when not specified nodespec by "-A" command. */
		if (i->hb_node_name == NULL || strlen(i->hb_node_name) == 0) {
			continue;
		}
		i->hbc_node = find_or_add_hbc_node(i->hb_node_name);
	}   

	return 0;
}

int free_server_conf(struct server_conf * conf)
{
	if(!conf)
		return 0;
	if(conf->iface)
		free(conf->iface);
	if(conf->hwaddr)
		free(conf->hwaddr);
	if(conf->riname)
		free(conf->riname);
	if(conf->ri_opt)
		free(conf->ri_opt);
	free(conf);
	return 0;
}

int free_server_conf_list(struct server_conf * info)
{
	struct server_conf * i, *j = NULL;
	if(info == NULL)
		return 0;
	for(i = info->next, j = info;
		i != NULL;
		j = i, i = j->next) {
		free_node_info(j->ni);
		free_server_conf(j);
	}
	return 0;
}

int serialize_server_conf(const struct server_conf * conf, char * str)
{
	struct options opt;
	ServerConf2opt(conf, &opt);
	opt2str(&opt,str, INNER_OPT_PARSE);
	return 0;
}

int deserialize_server_conf(char * str, struct server_conf * conf)
{
	struct options opt;
	str2opt(str, &opt, INNER_OPT_PARSE);
	opt2ServerConf(&opt, conf);
	return 0;
}

int ServerConf2str(const struct server_conf * conf, char * str, int maxlen)
{
	const struct server_conf * i;
	const struct dtsd_node_info * j;
	int rest = maxlen;

	if (conf == NULL) {
		strcpy(str, "");
		return 0;
	}
	for_each_server_conf(i, conf) {
		int len;
		char buffer[MAX_OPT_LEN];
		serialize_server_conf(i, buffer);
		len = strlen(buffer);
		rest -= len;
		if(rest < 0) {
			eprintf(HERE "failed to serialize server configuration, exceed a message size.\n");
			return -1;
		}
		snprintf(str, len + 1, "%s\n", buffer);
		str += len;
		for_each_node_info(j, i) {
			serialize_NodeInfo(j, buffer);
			len = strlen(buffer);
			rest -= len;
			if(rest < 0) {
				eprintf(HERE "failed to serialize server configuration, exceed a message size.\n");
				return -1;
			}
			snprintf(str, len + 1, "%s\n", buffer);
			str += len;
		}
	}
	debug_printf(HERE "succeeded to serialize server configuration.\n");
	return 0;
}

int str2ServerConf(const char * str, struct server_conf ** const conf)
{
	struct server_conf * i = NULL, * tmp_conf = NULL, * root;
	struct dtsd_node_info * j = NULL, * tmp_info = NULL;
	struct options opt;
	const char * head, * tail;

	root = NULL;
	head = str;
	if(conf == NULL) {
		eprintf(HERE "Invalid Arguments");
		return -1;
	}

	if (strlen(str) == 0) {
		*conf = NULL;
		return 0;
	}

	do{
		char * token = NULL;
		tail = strchr(head, '\n');
		if(tail == NULL) {
			if(*head == '\0')
				break;
			token = strdup(head);
			head = NULL;
		} else if (tail > head + 1) {
			token = strndup(head, tail - head);
			head = ++tail;
		} else {
			break;
		}

		debug_fprintf(stderr, HERE "parsing (\"%s\")\n", token);
		if(str2opt(token, &opt, INNER_OPT_PARSE) != 0) {
			eprintf(HERE "failed to deserialize options\n");
			goto free_all;
		}
		debug_fprintf(stderr, HERE "succeeded to deserialize options\n");
		if(opt.cmd == ADD_CONF) {
			tmp_conf = (struct server_conf *) malloc (sizeof(struct server_conf));
			opt2ServerConf(&opt, tmp_conf);
			if(root == NULL) {
				root = tmp_conf;
				i = tmp_conf;
			} else {
				i->next = tmp_conf;
				i = tmp_conf;
			}
			i->ni = NULL;
			j = NULL;

			if(strlen(opt.nodes) != 0) {
				char * nodes = strdup(opt.nodes);
				struct dtsd_node_info * info_list;
				if(parse_nodespec_list(nodes, &info_list) != 0) {
					eprintf(HERE "failed to parse node list\n");
					free(nodes);
					goto free_all;
				}
				free(nodes);
				add_NodeInfo(i, info_list);
			}
		} else if (opt.cmd == ADD_NODE) {
			tmp_info = (struct dtsd_node_info *) malloc (sizeof(struct dtsd_node_info));
			opt2NodeInfo(&opt, tmp_info);
			if(i->ni == NULL) {
				i->ni = tmp_info;
				j = i->ni;
			} else {
				j->next = tmp_info;
				j=j->next;
			}
			j->srv_conf = i;
		} else {
			free(token);
			goto free_all;
		}
		free(token);
	} while(head != NULL);

	assert(conf != NULL);
	*conf = root;

	return 0;

free_all:
	if(root != NULL)
		free_server_conf_list(root);
	return 1;
}

int count_opt(const struct server_conf * server_conf) {
	const struct server_conf * i;
	const struct dtsd_node_info * j;
	int num = 0;
	if(server_conf == NULL)
		return num;
	for_each_server_conf(i, server_conf) {
		num++;
		for_each_node_info(j, i) {
			num++;
		}
	}
	return num;
}

int apply_ServerConf(struct dtsd_config * dtsd_conf, const struct server_conf * conf) {
	const struct server_conf * i;
	const struct dtsd_node_info * j;
	char res[MAX_CMD_RES_SIZE] = "";
	char err[MAX_CMD_RES_SIZE] = "";

	if(conf == NULL) {
		debug_printf(HERE "No Server Configuration.\n");
		return 0;
	}
	for_each_server_conf(i, conf) {
		if(bind_socket(i->proxy_port) != 0) {
			eprintf(HERE "failed to bind socket with port %d.\n", i->proxy_port);
			continue;
		} else {
			debug_printf(HERE "succeeded to bind socket with port %d.\n", i->proxy_port);
		}
		if(exec_ServerConf(i, dtsd_conf, res, err) != 0) {
			eprintf(HERE "failed to exec server conf.");
			return -1;
		}
		// initialize and run the ri.
		if (ri_intf_init_run(dtsd_conf, i) != 0) {
			debug_printf(HERE "failed to initialize and run the Request Inspector.\n");
			strcpy(err, "failed to initialize and run the Request Inspector.\n");
			return -1;
		}
		for_each_node_info(j, i) {
			if(exec_NodeInfo(j, i, res, err)) {
				continue;
			}
		}
	}
	dtsd_conf->server_conf = (struct server_conf *) conf;
	return 0;
}

int add_NodeInfo(struct server_conf * conf, struct dtsd_node_info * info)
{
	struct dtsd_node_info * i, * j;
	if(conf == NULL) {
		eprintf(HERE "server_conf is not specified\n");
		return -1;
	}
	if(info == NULL) {
		eprintf(HERE "dtsd_node_info is not specified\n");
		return -1;
	}

	if(conf->ni == NULL) {
		conf->ni = info;
		for_each_node_info(j, conf) {
			j->srv_conf = conf;
		}
		return 0;
	}

	for_each_node_info(i, conf) {
		if(strcmp(i->hb_node_name,info->hb_node_name) == 0) {
			eprintf(HERE "intra_addr duplicated.\n");
			return -1;
		}
		if(i->next == NULL) {
			i->next = info;
			i = info;
			while(i != NULL) {
				i->srv_conf = conf;
				i = i->next;
			}
			return 0;
		}
	}
	return -1;
}

int remove_NodeInfo(struct server_conf * conf, const char * hb_node_name)
{
	struct dtsd_node_info * i, *j;
	if(conf == NULL) {
		eprintf(HERE "server_conf not specified\n");
		return -1;
	}
	if(conf->ni == NULL) {
		debug_printf(HERE "no node named %s\n", hb_node_name);
		return 0;
	}
	if(strcmp(conf->ni->hb_node_name, hb_node_name) == 0) {
		i = conf->ni;
		conf->ni = i->next;
		free_node_info(i);
	}
	for_each_node_info(i, conf) {
		if(i->next == NULL)
			break;
		if(strcmp(i->next->hb_node_name, hb_node_name) == 0) {
			j = i->next;
			i->next = j->next;
			free_node_info(j);
			return 0;
		}
	}
	debug_printf(HERE "no node named %s\n", hb_node_name);
	return 0;
}

int append_ServerConfList(struct server_conf * list, struct server_conf * conf)
{
	struct server_conf * i;
	if(list == NULL) {
		eprintf(HERE "server_conf root is not specified\n");
		return -1;
	}
	if(conf == NULL) {
		eprintf(HERE "server_conf is not specified\n");
		return -1;
	}
	for_each_server_conf(i, list) {
		if(i->server_address == conf->server_address &&
		   i->server_port    == conf->server_port) {
			eprintf(HERE "server address and port is duplicated.\n");
			return -1;
		}
		if(i->next == NULL) {
			i->next = conf;
			break;
		}
	}
	return 0;
}

struct server_conf * find_ServerConf(const struct dtsd_config * dconf,
									 const char * host_srv)
{
	uint32_t addr;
	int port;
	struct server_conf * i;
	if(parse_hostsrv(host_srv, &addr, &port) != 0) {
		eprintf(HERE "Invalid host:srv %s\n", host_srv);
		return NULL;
	}
	for_each_server_conf(i, dconf->server_conf) {
		if(i->server_address == addr &&
		   i->server_port    == port) {
			return i;
		}
	}
	return NULL;
}

int add_ServerConf(struct dtsd_config * dconf, struct server_conf * sconf)
{
	if(dconf->server_conf == NULL) {
		dconf->server_conf = sconf;
		return 0;
	} else {
		return append_ServerConfList(dconf->server_conf, sconf);
	}
}

int append_OptionsAsServerConf(struct server_conf ** conf, struct options * opt)
{
	if(opt->cmd == ADD_CONF) {
		struct server_conf * nconf;
		nconf = (struct server_conf *) malloc ((sizeof(struct server_conf)));
		if(opt2ServerConf(opt, nconf) != 0) {
			free(nconf);
			return -1;
		}
		if(*conf == NULL) {
			*conf = nconf;
			return 0;
		}
		if(append_ServerConfList(*conf, nconf) != 0) {
			free(nconf);
			return -1;
		}
		return 0;

	} else if (opt->cmd == ADD_NODE) {
		struct server_conf * i;
		struct dtsd_node_info * ninfo;
		uint32_t server_addr;
		int port;
		if(*conf == NULL) {
			return -1;
		}
		if(parse_hostsrv(opt->host_srv, &server_addr, &port)) {
			return -1;
		}
		ninfo = (struct dtsd_node_info *) malloc ((sizeof(struct dtsd_node_info)));
		opt2NodeInfo(opt, ninfo);
		for_each_server_conf(i, *conf) {
			if(i->server_address == server_addr &&
			   i->server_port    == port) {
				if(add_NodeInfo(i, ninfo) != 0) {
					free(ninfo);
					return -1;
				}
			}
		}
		return 0;
	} else {
		eprintf(HERE "opt is neither ADD_CONF or ADD_NODE\n");
		return -1;
	}
}
