#include <cutter.h>
#include <stdio.h>
#include <string.h>
#include <sys/queue.h>

#include "dtsd_optparser.h"
#include "dtsadmin_result_handler.h"
#include "dtsadmin_server_conf_handler.h"
#include "dtsadmin_nodeinfo_handler.h"

/* start the dtsadmin's test codes. */
extern int decide_proxy_port(const struct options * opt);
extern int sconf2AddNodeCmdStrv(const struct server_confs * sconf, char ** opt_strv[], int * cnt);
extern int sconf2NodeInfoStrv(const struct server_confs * sconf, int nodeWithHash, char ** list_strv[], int * cnt);

void test_decide_proxy_port()
{
	struct options opt
		= { 0, 0, "192.168.10.1:8080", "", "", "", 0, "", "", "", 0, 0, 0, 0, 0 };
	int ret;

	ret = decide_proxy_port(&opt);
	cut_assert_equal_int(53272, ret);

	strcpy(opt.host_srv, "192.168.10.1:30000");
	ret = decide_proxy_port(&opt);
	cut_assert_equal_int(54849, ret);

	strcpy(opt.host_srv, "192.168.10.1:0");
	ret = decide_proxy_port(&opt);
	cut_assert_equal_int(45192, ret);

	strcpy(opt.host_srv, "192.168.10.1:65535");
	ret = decide_proxy_port(&opt);
	cut_assert_equal_int(49698, ret);
}

void test_sconf2AddNodeCmdStrv()
{
	struct server_confs_head shead = TAILQ_HEAD_INITIALIZER(shead);
	struct server_confs * sconf;

	int i, ret, cnt;
	char ** opt_strv;
	const char msg_1[] =
"?srv_addr=192.168.30.1?srv_port=80?iface=eth0\
?hwaddr=aa:bb:cc:dd:ee:ff?netmask=255.255.255.0?fakemac=on\
?ri=httpri?ri_opt=rr\?proxy_port=53272\
?node=dts01%10.0.0.1/1#00000001:dts02%10.0.0.2/2#00000002\n\
?srv_addr=192.168.30.1?srv_port=22?iface=eth0\
?hwaddr=ff:ee:dd:cc:bb:aa?netmask=255.255.254.0?fakemac=off\
?ri=ftpri?ri_opt=rr\?proxy_port=45214\
?node=dts03%10.0.0.3/3#00000003:dts04%10.0.0.4/4#00000004\n";

	ret = str2sconfq(msg_1, &shead);
	cut_assert_equal_int(0, ret);

	sconf = TAILQ_FIRST(&shead);
	ret = sconf2AddNodeCmdStrv(sconf, &opt_strv, &cnt);

	cut_assert_equal_int(0, ret);
	cut_assert_equal_int(2, cnt);

	cut_assert_equal_string("-a 192.168.30.1:80 dts01/1", opt_strv[0]);
	cut_assert_equal_string("-a 192.168.30.1:80 dts02/2", opt_strv[1]);

	/* free the opt_strv */
	for (i = 0; i < cnt; i++) {
		free(opt_strv[i]);
	}
	free(opt_strv);
	TAILQ_REMOVE(&shead, sconf, entry);

	sconf = TAILQ_FIRST(&shead);
	ret = sconf2AddNodeCmdStrv(sconf, &opt_strv, &cnt);
	cut_assert_equal_int(0, ret);
	cut_assert_equal_int(2, cnt);

	cut_assert_equal_string("-a 192.168.30.1:22 dts03/3", opt_strv[0]);
	cut_assert_equal_string("-a 192.168.30.1:22 dts04/4", opt_strv[1]);

	/* free the opt_strv */
	for (i = 0; i < cnt; i++) {
		free(opt_strv[i]);
	}
	free(opt_strv);
}

void test_sconf2NodeInfoStrv()
{
	struct server_confs_head shead = TAILQ_HEAD_INITIALIZER(shead);
	struct server_confs * sconf;

	int i, ret, cnt;
	char ** opt_strv;
	const char msg_1[] =
"?srv_addr=192.168.30.1?srv_port=80?iface=eth0\
?hwaddr=aa:bb:cc:dd:ee:ff?netmask=255.255.255.0?fakemac=on\
?ri=httpri?ri_opt=rr\?proxy_port=53272\
?node=dts01%10.0.0.1/1#00000001:dts02%10.0.0.2/2#00000002\n\
?srv_addr=192.168.30.1?srv_port=22?iface=eth0\
?hwaddr=ff:ee:dd:cc:bb:aa?netmask=255.255.254.0?fakemac=off\
?ri=ftpri?ri_opt=rr\?proxy_port=45214\
?node=dts03%10.0.0.3/3#00000003:dts04%10.0.0.4/4#00000004\n";

	ret = str2sconfq(msg_1, &shead);
	cut_assert_equal_int(0, ret);

	sconf = TAILQ_FIRST(&shead);
	ret = sconf2NodeInfoStrv(sconf, 1, &opt_strv, &cnt);
	cut_assert_equal_int(0, ret);
	cut_assert_equal_int(2, cnt);

	cut_assert_equal_string("  -> dts01 1 00000001", opt_strv[0]);
	cut_assert_equal_string("  -> dts02 2 00000002", opt_strv[1]);

	/* free the opt_strv */
	for (i = 0; i < cnt; i++) {
		free(opt_strv[i]);
	}
	free(opt_strv);
	TAILQ_REMOVE(&shead, sconf, entry);

	sconf = TAILQ_FIRST(&shead);
	ret = sconf2NodeInfoStrv(sconf, 1, &opt_strv, &cnt);
	cut_assert_equal_int(0, ret);
	cut_assert_equal_int(2, cnt);

	cut_assert_equal_string("  -> dts03 3 00000003", opt_strv[0]);
	cut_assert_equal_string("  -> dts04 4 00000004", opt_strv[1]);

	/* free the opt_strv */
	for (i = 0; i < cnt; i++) {
		free(opt_strv[i]);
	}
	free(opt_strv);
	TAILQ_REMOVE(&shead, sconf, entry);
}

/* start the dtsadmin_node_info_handler's test codes. */
void assert_ninfo(struct node_info * expected, struct node_info * actual)
{
	cut_assert_equal_string(expected->hb_node_name, actual->hb_node_name);
	cut_assert_equal_string(expected->addr, actual->addr);
	cut_assert_equal_string(expected->id, actual->id);
	cut_assert_equal_string(expected->id_hash, actual->id_hash);
}

void test_str2ninfoq()
{
	int ret;

	/* head of result queue. */
	struct node_info_head nhead = TAILQ_HEAD_INITIALIZER(nhead);
	struct node_info * ninfo;

	/* set value. */
	const char msg_1[] = "dts01%10.0.0.1/1#00000001:dts02%10.0.0.2/2#00000002:";

	ret = str2ninfoq(msg_1, &nhead);
	cut_assert_equal_int(0, ret);

	/* deque and assert. */
	struct node_info expected_1
			= { "dts01", "10.0.0.1", "1", "00000001" };
	ninfo = TAILQ_FIRST(&nhead);
	cut_assert_not_null(ninfo);
	assert_ninfo(&expected_1, ninfo);
	TAILQ_REMOVE(&nhead, ninfo, entry);

	struct node_info expected_2
			= { "dts02", "10.0.0.2", "2", "00000002" };
	ninfo = TAILQ_FIRST(&nhead);
	cut_assert_not_null(ninfo);
	assert_ninfo(&expected_2, ninfo);
	TAILQ_REMOVE(&nhead, ninfo, entry);

	/* invalid format test */
	const char msg_2[] = "not null";
	struct node_info expected_3
			= { "not null", "", "", "" };

	ret = str2ninfoq(msg_2, &nhead);
	cut_assert_equal_int(0, ret);
	/* if #, /, : is not exist, the msg is treated as hb_node_name.. */
	cut_assert_equal_int(0, ret);
	ninfo = TAILQ_FIRST(&nhead);
	cut_assert_not_null(ninfo);
	assert_ninfo(&expected_3, ninfo);
	TAILQ_REMOVE(&nhead, ninfo, entry);
}

void test_free_ninfoq()
{
	int ret;

	struct node_info_head nhead= TAILQ_HEAD_INITIALIZER(nhead);
	struct node_info * ninfo;

	const char msg_1[] = "dts01%10.0.0.1/1#00000001:dts02%10.0.0.2/2#00000002";

	ret = str2ninfoq(msg_1, &nhead);
	cut_assert_equal_int(0, ret);
	ninfo = TAILQ_FIRST(&nhead);
	cut_assert_not_null(ninfo);

	free_ninfoq(&nhead);
	ninfo = TAILQ_FIRST(&nhead);
	cut_assert_null(ninfo);
}

/* start the dtsadmin_server_conf_handler's test codes. */
void assert_sconf(struct server_confs * expected, struct server_confs * actual)
{
	cut_assert_equal_string(expected->srv_addr, actual->srv_addr);
	cut_assert_equal_string(expected->srv_port, actual->srv_port);
	cut_assert_equal_string(expected->iface, actual->iface);
	cut_assert_equal_string(expected->macaddr, actual->macaddr);
	cut_assert_equal_string(expected->netmask, actual->netmask);
	cut_assert_equal_string(expected->fakemac, actual->fakemac);
	cut_assert_equal_string(expected->ri, actual->ri);
	cut_assert_equal_string(expected->ri_opt, actual->ri_opt);
	cut_assert_equal_string(expected->proxy_port, actual->proxy_port);
	cut_assert_equal_string(expected->nodes, actual->nodes);
}

void test_str2sconfq()
{
	int ret;

	struct server_confs_head shead = TAILQ_HEAD_INITIALIZER(shead);
	struct server_confs * sconf;

	const char msg_1[] =
"?srv_addr=192.168.30.1?srv_port=80?iface=eth0\
?hwaddr=aa:bb:cc:dd:ee:ff?netmask=255.255.255.0?fakemac=on\
?ri=httpri?ri_opt=rr\?proxy_port=53272\
?node=dts01/1#00000001:dts02/2#00000002\n\
?srv_addr=192.168.30.1?srv_port=22?iface=eth0\
?hwaddr=ff:ee:dd:cc:bb:aa?netmask=255.255.254.0?fakemac=off\
?ri=ftpri?ri_opt=rr\?proxy_port=45214\
?node=dts03/3#00000003:dts04/4#00000004\n";

	ret = str2sconfq(msg_1, &shead);
	cut_assert_equal_int(0, ret);

	struct server_confs expected_1
			= { "192.168.30.1", "80", "eth0", "aa:bb:cc:dd:ee:ff",
				"255.255.255.0", "on", "httpri", "rr", "53272",
				"dts01/1#00000001:dts02/2#00000002" };
	sconf = TAILQ_FIRST(&shead);
	cut_assert_not_null(sconf);
	assert_sconf(&expected_1, sconf);
	TAILQ_REMOVE(&shead, sconf, entry);

	struct server_confs expected_2
			= { "192.168.30.1", "22", "eth0", "ff:ee:dd:cc:bb:aa",
				"255.255.254.0", "off", "ftpri", "rr", "45214",
				"dts03/3#00000003:dts04/4#00000004" };
	sconf = TAILQ_FIRST(&shead);
	cut_assert_not_null(sconf);
	assert_sconf(&expected_2, sconf);
	TAILQ_REMOVE(&shead, sconf, entry);

	/* invalid format test */
	const char msg_2[] =
"?=192.168.30.1?srv_port=80?iface=eth0\
?hwaddr=aa:bb:cc:dd:ee:ff?netmask=255.255.255.0?fakemac=on\
?ri=httpri?ri_opt=rr\?proxy_port=53272\
?node=dts01/1#00000001:dts02/2#00000002\n\
?=192.168.30.1?srv_port=22?iface=eth0\
?hwaddr=ff:ee:dd:cc:bb:aa?netmask=255.255.254.0?fakemac=off\
?ri=ftpri?ri_opt=rr\?proxy_port=45214\
?node=dts03/3#00000003:dts04/4#00000004\n";

	ret = str2sconfq(msg_2, &shead);
	/* if ?srv_addr not exist, nothing to add and return 0.  */
	cut_assert_equal_int(0, ret);
	sconf = TAILQ_FIRST(&shead);
	cut_assert_null(sconf);
}

void test_free_sconfq()
{
	int ret;

	/* head of result queue. */
	struct server_confs_head shead = TAILQ_HEAD_INITIALIZER(shead);
	struct server_confs * sconf;

	/* set value. */
	const char msg_1[] =
"?srv_addr=192.168.30.1?srv_port=80?iface=eth0\
?hwaddr=aa:bb:cc:dd:ee:ff?netmask=255.255.255.0?fakemac=on\
?ri=httpri?ri_opt=rr\?proxy_port=53272\
?node=dts01/1#00000001:dts02/2#00000002\n\
?srv_addr=192.168.30.1?srv_port=22?iface=eth0\
?hwaddr=ff:ee:dd:cc:bb:aa?netmask=255.255.254.0?fakemac=off\
?ri=ftpri?ri_opt=rr\?proxy_port=45214\
?node=dts03/3#00000003:dts04/4#00000004\n";
	ret = str2sconfq(msg_1, &shead);
	cut_assert_equal_int(0, ret);
	sconf = TAILQ_FIRST(&shead);
	cut_assert_not_null(sconf);

	free_sconfq(&shead);
	sconf = TAILQ_FIRST(&shead);
	cut_assert_null(sconf);
}

/* start the dtsadmin_result_handler's test codes. */
void set_resultarr_correct_format(char * strarr[], const int result_cnt)
{
	int i;
	char msg[1024];

	for (i = 0; i < result_cnt; i++) {
		strarr[i] = (char *) malloc(sizeof(char) * 1024);
		snprintf(msg, 1024, ">>\nHOST-host%d\nRESULT-OK\nERR-noerr\nMSG-nomsg\n", i);
		strncpy(strarr[i], msg, 1024);
	}
}

/* start the dtsadmin_result_handler's test codes. */
void set_resultarr_invalid_format(char * strarr[], const int result_cnt)
{
	int i;
	char msg[1024];

	for (i = 0; i < result_cnt; i++) {
		strarr[i] = (char *) malloc(sizeof(char) * 1024);
		snprintf(msg, 1024, ">>\nHOST-host%d\nERR-noerr\nMSG-nomsg\n", i);
		strncpy(strarr[i], msg, 1024);
	}
}

/* start the dtsadmin_result_handler's test codes. */
void set_resultarr_exist_stat_ng(char * strarr[], const int result_cnt)
{
	int i;
	char msg[1024];

	for (i = 0; i < result_cnt; i++) {
		strarr[i] = (char *) malloc(sizeof(char) * 1024);
		if (i == 0) {
			snprintf(msg, 1024, ">>\nHOST-host%d\nRESULT-NG\nERR-noerr\nMSG-nomsg\n", i);
		} else {
			snprintf(msg, 1024, ">>\nHOST-host%d\nRESULT-OK\nERR-noerr\nMSG-nomsg\n", i);
		}
		strncpy(strarr[i], msg, 1024);
	}
}

/* the strarr's each message is correct format. */
void test_strarr2resultq()
{
	int result_cnt = 3;
	int i, ret;
	char *strarr[result_cnt];
	/* head of result queue. */
	struct results_head rhead = TAILQ_HEAD_INITIALIZER(rhead);

	/* set value. */
	set_resultarr_correct_format(strarr, result_cnt);
	/* convert results and insert to queue. */
	ret = strarr2resultq((const char **) strarr, result_cnt, &rhead);
	cut_assert_equal_int(0, ret);

	/* deque and assert. */
	struct processed_results * result;

	char expected_host[100];
	for (i = 0; i < result_cnt; i++) {
		result = TAILQ_FIRST(&rhead);
		cut_assert_not_null(result);
		snprintf(expected_host, 100, "host%d", i);
		cut_assert_equal_string(expected_host, result->host);
		cut_assert_equal_string("OK", result->stat);
		cut_assert_equal_string("noerr", result->error);
		cut_assert_equal_string("nomsg", result->msg);
		TAILQ_REMOVE(&rhead, result, entry);
	}

	/* invalid format test */
	set_resultarr_invalid_format(strarr, result_cnt);
	ret = strarr2resultq((const char **) strarr, result_cnt, &rhead);
	cut_assert_equal_int(-1, ret);

	/* deque and assert. */
	result = TAILQ_FIRST(&rhead);
	cut_assert_null(result);
}

/* the strarr's each message is correct format. */
void test_free_resultq()
{
	int result_cnt = 3;
	int i, ret;
	char *strarr[result_cnt];
	/* head of result queue. */
	struct results_head rhead = TAILQ_HEAD_INITIALIZER(rhead);

	/* set value. */
	set_resultarr_correct_format(strarr, result_cnt);
	/* convert results and insert to queue. */
	ret = strarr2resultq((const char **) strarr, result_cnt, &rhead);
	cut_assert_equal_int(0, ret);

	/* free the result queue. */
	free_resultq(&rhead);

	struct processed_results * result;
	result = TAILQ_FIRST(&rhead);
	cut_assert_null(result);

	/* re-free result queue. */
	cut_assert_null(result);
}

/* the strarr's each message is correct format. */
void test_judge_resultq()
{
	int result_cnt = 3;
	int i, ret;
	char *strarr[result_cnt];
	/* head of result queue. */
	struct results_head rhead = TAILQ_HEAD_INITIALIZER(rhead);

	/* set value. */
	set_resultarr_correct_format(strarr, result_cnt);
	/* convert results and insert to queue. */
	ret = strarr2resultq((const char **) strarr, result_cnt, &rhead);
	cut_assert_equal_int(0, ret);
	ret = judge_resultq(&rhead);
	cut_assert_equal_int(0, ret);

	/* result->stat = "NG". */
	set_resultarr_exist_stat_ng(strarr, result_cnt);
	ret = strarr2resultq((const char **) strarr, result_cnt, &rhead);
	cut_assert_equal_int(0, ret);
	ret = judge_resultq(&rhead);
	cut_assert_equal_int(-1, ret);
}

/* start the dtsd_optparser's test codes. */
void clear_opt(struct options * opt)
{
	opt->cmd = 0;
	strcpy(opt->host_srv, "");
	strcpy(opt->iface, "");
	strcpy(opt->netmask, "");
	strcpy(opt->macaddr, "");
	opt->fakemac = 0;
	strcpy(opt->ri, "");
	strcpy(opt->ri_opt, "");
	strcpy(opt->nodes, "");
	opt->proxy_port = 0;
	opt->chash = 0;
	opt->dead_node = 0;
	opt->is_inner_cmd = 0;
}

void assert_opt(struct options * expected, struct options * actual)
{
	cut_assert_equal_int(expected->cmd, actual->cmd);
	cut_assert_equal_string(expected->host_srv, actual->host_srv);
	cut_assert_equal_string(expected->iface, actual->iface);
	cut_assert_equal_string(expected->netmask, actual->netmask);
	cut_assert_equal_string(expected->macaddr, actual->macaddr);
	cut_assert_equal_int(expected->fakemac, actual->fakemac);
	cut_assert_equal_string(expected->ri, actual->ri);
	cut_assert_equal_string(expected->ri_opt, actual->ri_opt);
	cut_assert_equal_string(expected->nodes, actual->nodes);
	cut_assert_equal_int(expected->proxy_port, actual->proxy_port);
	cut_assert_equal_int(expected->chash, actual->chash);
	cut_assert_equal_int(expected->dead_node, actual->dead_node);
	cut_assert_equal_int(expected->is_inner_cmd, actual->is_inner_cmd);
}

void test_str2opt_opt2str()
{
	char opt2strs_str[4096];
	struct options opt_1
		= { 0, 0, "", "", "", "", 0, "", "", "", 0, 0, 0, 0, 0 };
	struct options opt_2
		= { 0, 0, "", "", "", "", 0, "", "", "", 0, 0, 0, 0, 0 };
	int ret;

	/* no netmask */
	char opt_str_1[] = "-A 192.168.1.1:80 -i eth0 -f on -r httpri -s \"rr test\" 10.0.0.1:10.0.0.2\n";
	ret = str2opt(opt_str_1, &opt_1, OUTER_OPT_PARSE);
	struct options expected_1
		= { ADD_CONF, 0, "192.168.1.1:80", "eth0", "255.255.255.0", "", 1, "httpri", "rr test", "10.0.0.1:10.0.0.2", 0, 0, 0, 0 };
	cut_assert_equal_int(0, ret);
	assert_opt(&expected_1, &opt_1);

	ret = opt2str((const struct options *) &opt_1, opt2strs_str, OUTER_OPT_PARSE);
	cut_assert_equal_int(0, ret);

	ret = str2opt((const char *) &opt2strs_str, &opt_2, OUTER_OPT_PARSE);
	cut_assert_equal_int(0, ret);

	assert_opt(&opt_1, &opt_2);
}

void test_str2opt_outer_parse_A()
{
	struct options opt
		= { 0, 0, "", "", "", "", 0, "", "", "", 0, 0, 0, 0, 0 };
	int ret;

	/* no netmask */
	char opt_str_1[] = "-A 192.168.1.1:80 -i eth0 -f on -r httpri -s \"rr test\" 10.0.0.1:10.0.0.2\n";
	ret = str2opt(opt_str_1, &opt, OUTER_OPT_PARSE);
	struct options expected_1
		= { ADD_CONF, 0, "192.168.1.1:80", "eth0", "255.255.255.0", "", 1, "httpri", "rr test", "10.0.0.1:10.0.0.2", 0, 0, 0, 0, 0 };
	cut_assert_equal_int(0, ret);
	assert_opt(&expected_1, &opt);
	clear_opt(&opt);

	/* with netmask */
	char opt_str_2[] = "-A 192.168.1.1:80 -i eth0 -f on -r httpri -M 255.255.254.0 -s \"rr test\" 10.0.0.1:10.0.0.2";
	ret = str2opt(opt_str_2, &opt, OUTER_OPT_PARSE);
	struct options expected_2
		= { ADD_CONF, 0, "192.168.1.1:80", "eth0", "255.255.254.0", "", 1, "httpri", "rr test", "10.0.0.1:10.0.0.2", 0, 0, 0, 0, 0 };
	cut_assert_equal_int(0, ret);
	assert_opt(&expected_2, &opt);
	clear_opt(&opt);

	/* without nodes */
	char opt_str_3[] = "-A 192.168.1.1:80 -i eth0 -f on -r httpri -M 255.255.254.0 -s \"rr test\"";
	ret = str2opt(opt_str_3, &opt, OUTER_OPT_PARSE);
	struct options expected_3
		= { ADD_CONF, 0, "192.168.1.1:80", "eth0", "255.255.254.0", "", 1, "httpri", "rr test", "", 0, 0, 0, 0, 0 };
	cut_assert_equal_int(0, ret);
	assert_opt(&expected_3, &opt);
}

void test_opt2str_outer_parse_A()
{
	struct options opt
		= { 0, 0, "", "", "", "", 0, "", "", "", 0, 0, 0, 0, 0 };
	int ret;
	char str[65535];

	char opt_str_1[] = "-A 192.168.1.1:80 -i eth0 -f on -r httpri -s \"rr test\" 10.0.0.1:10.0.0.2";
	char expected_str_1[] = "-A 192.168.1.1:80 -i eth0 -M 255.255.255.0   -f on -r httpri -s \"rr test\" 10.0.0.1:10.0.0.2\n";
	ret = str2opt(opt_str_1, &opt, OUTER_OPT_PARSE);
	ret = opt2str(&opt, str, OUTER_OPT_PARSE);
	cut_assert_equal_int(0, ret);
	cut_assert_equal_string(expected_str_1, str);
	clear_opt(&opt);

	/* without nodes */
	char opt_str_2[] = "-A 192.168.1.1:80 -i eth0 -f on -r httpri -s \"rr test\"";
	char expected_str_2[] = "-A 192.168.1.1:80 -i eth0 -M 255.255.255.0   -f on -r httpri -s \"rr test\" \n";
	ret = str2opt(opt_str_2, &opt, OUTER_OPT_PARSE);
	ret = opt2str(&opt, str, OUTER_OPT_PARSE);
	cut_assert_equal_int(0, ret);
	cut_assert_equal_string(expected_str_2, str);
}

void test_str2opt_outer_parse_a()
{
	struct options opt
		= { 0, 0, "", "", "", "", 0, "", "", "", 0, 0, 0, 0, 0 };
	int ret;

	char opt_str_1[] = "-a 192.168.1.1:80 dts01/1-3";
	ret = str2opt(opt_str_1, &opt, OUTER_OPT_PARSE);
	struct options expected_1
		= { ADD_NODE, 0, "192.168.1.1:80", "", "", "", 0, "", "", "dts01/1-3", 0, 0, 0, 0, 0 };
	cut_assert_equal_int(0, ret);
	assert_opt(&expected_1, &opt);
}

void test_opt2str_outer_parse_a()
{
	struct options opt
		= { 0, 0, "", "", "", "", 0, "", "", "", 0, 0, 0, 0, 0 };
	int ret;
	char str[65535];

	char opt_str_1[] = "-a 192.168.1.1:80 dts01/1-3";
	str2opt(opt_str_1, &opt, OUTER_OPT_PARSE);
	char expected_str_1[] = "-a \"192.168.1.1:80\" \"dts01/1-3\"\n";
	ret = str2opt(opt_str_1, &opt, OUTER_OPT_PARSE);
	ret = opt2str(&opt, str, OUTER_OPT_PARSE);
	cut_assert_equal_int(0, ret);
	cut_assert_equal_string(expected_str_1, str);
}

void test_str2opt_outer_parse_d()
{
	struct options opt
		= { 0, 0, "", "", "", "", 0, "", "", "", 0, 0, 0, 0, 0 };
	int ret;

	char opt_str_1[] = "-d 192.168.1.1:80 dts01/1-3";
	ret = str2opt(opt_str_1, &opt, OUTER_OPT_PARSE);
	struct options expected_1
		= { DEL_NODE, 0, "192.168.1.1:80", "", "", "", 0, "", "", "dts01/1-3", 0, 0, 0, 0, 0 };
	cut_assert_equal_int(0, ret);
	assert_opt(&expected_1, &opt);
}

void test_opt2str_outer_parse_d()
{
	struct options opt
		= { 0, 0, "", "", "", "", 0, "", "", "", 0, 0, 0, 0, 0 };
	int ret;
	char str[65535];

	char opt_str_1[] = "-d 192.168.1.1:80 dts01/1-3";
	str2opt(opt_str_1, &opt, OUTER_OPT_PARSE);
	char expected_str_1[] = "-d \"192.168.1.1:80\" \"dts01/1-3\"\n";
	ret = str2opt(opt_str_1, &opt, OUTER_OPT_PARSE);
	ret = opt2str(&opt, str, OUTER_OPT_PARSE);
	cut_assert_equal_int(0, ret);
	cut_assert_equal_string(expected_str_1, str);
}

void test_str2opt_outer_parse_L()
{
	struct options opt
		= { 0, 0, "", "", "", "", 0, "", "", "", 0, 0, 0, 0, 0 };
	int ret;

	/* only -L */
	char opt_str_1[] = "-L";
	ret = str2opt(opt_str_1, &opt, OUTER_OPT_PARSE);
	struct options expected_1
		= { LIST, 0, "", "", "", "", 0, "", "", "", 0, 0, 0, 0, 0 };
	cut_assert_equal_int(0, ret);
	assert_opt(&expected_1, &opt);
	clear_opt(&opt);

	/* with -n */
	char opt_str_2[] = "-L -n";
	ret = str2opt(opt_str_2, &opt, OUTER_OPT_PARSE);
	struct options expected_2
		= { LIST, 0, "", "", "", "", 0, "", "", "", 0, 0, 1, 0, 0 };
	cut_assert_equal_int(0, ret);
	assert_opt(&expected_2, &opt);
	clear_opt(&opt);

	/* with -c */
	char opt_str_3[] = "-L -c";
	ret = str2opt(opt_str_3, &opt, OUTER_OPT_PARSE);
	struct options expected_3
		= { LIST, 0, "", "", "", "", 0, "", "", "", 0, 1, 0, 0, 0 };
	cut_assert_equal_int(0, ret);
	assert_opt(&expected_3, &opt);
	clear_opt(&opt);

	/* -n and -c */
	char opt_str_4[] = "-L -n -c";
	ret = str2opt(opt_str_4, &opt, OUTER_OPT_PARSE);
	struct options expected_4
		= { LIST, 0, "", "", "", "", 0, "", "", "", 0, 1, 1, 0, 0 };
	cut_assert_equal_int(0, ret);
	assert_opt(&expected_4, &opt);
}

void test_opt2str_outer_parse_L()
{
	struct options opt
		= { 0, 0, "", "", "", "", 0, "", "", "", 0, 0, 0, 0, 0 };
	int ret;
	char str[65535];

	char opt_str_1[] = "-L";
	str2opt(opt_str_1, &opt, OUTER_OPT_PARSE);
	char expected_str_1[] = "-L  \n";
	ret = str2opt(opt_str_1, &opt, OUTER_OPT_PARSE);
	ret = opt2str(&opt, str, OUTER_OPT_PARSE);
	cut_assert_equal_int(0, ret);
	cut_assert_equal_string(expected_str_1, str);
	clear_opt(&opt);

	char opt_str_2[] = "-L -n";
	str2opt(opt_str_2, &opt, OUTER_OPT_PARSE);
	char expected_str_2[] = "-L  -n\n";
	ret = str2opt(opt_str_2, &opt, OUTER_OPT_PARSE);
	ret = opt2str(&opt, str, OUTER_OPT_PARSE);
	cut_assert_equal_int(0, ret);
	cut_assert_equal_string(expected_str_2, str);
	clear_opt(&opt);

	/* TODO -n -c */
}

void test_str2opt_inner_parse_A()
{
	struct options opt
		= { 0, 0, "", "", "", "", 0, "", "", "", 0, 0, 0, 0, 0 };
	int ret;

	char opt_str_1[] = "-A 192.168.0.30:9999 -i eth0 -M 255.255.255.0 -m a4:ba:db:03:7e:7e -p 45199 -f on -r httpri -s \"-d rr\" 10.0.0.1/1 ";
	ret = str2opt(opt_str_1, &opt, INNER_OPT_PARSE);
	struct options expected_1
		= { ADD_CONF, 0, "192.168.0.30:9999", "eth0", "255.255.255.0", "a4:ba:db:03:7e:7e", 1, "httpri", "-d rr", "10.0.0.1/1", 45199, 0, 0, 0, 0 };
	cut_assert_equal_int(0, ret);
	assert_opt(&expected_1, &opt);
	clear_opt(&opt);

	/* without nodes */
	char opt_str_2[] = "-A 192.168.1.1:80 -i eth0 -f on -m \"aa:bb:cc:dd:ee:ff\" -r httpri -p 65535 -s \"rr test\"";
	ret = str2opt(opt_str_2, &opt, INNER_OPT_PARSE);
	struct options expected_2
		= { ADD_CONF, 0, "192.168.1.1:80", "eth0", "255.255.255.0", "aa:bb:cc:dd:ee:ff", 1, "httpri", "rr test", "", 65535, 0, 0, 0, 0 };
	cut_assert_equal_int(0, ret);
	assert_opt(&expected_2, &opt);
}

void test_opt2str_inner_parse_A()
{
	struct options opt
		= { 0, 0, "", "", "", "", 0, "", "", "", 0, 0, 0, 0, 0 };
	int ret;
	char str[65535];

	char opt_str_1[] = "-A 192.168.1.1:80 -m aa:bb:cc:dd:ee:ff -p 65535 -i eth0 -f on -r httpri -s \"rr test\" 10.0.0.1:10.0.0.2";
	char expected_str_1[] = "-A 192.168.1.1:80 -i eth0 -M 255.255.255.0 -m aa:bb:cc:dd:ee:ff -p 65535 -f on -r httpri -s \"rr test\" 10.0.0.1:10.0.0.2\n";
	ret = str2opt(opt_str_1, &opt, INNER_OPT_PARSE);
	ret = opt2str(&opt, str, INNER_OPT_PARSE);
	cut_assert_equal_int(0, ret);
	cut_assert_equal_string(expected_str_1, str);
	clear_opt(&opt);

	/* without nodes */
	char opt_str_2[] = "-A 192.168.1.1:80 -m aa:bb:cc:dd:ee:ff -p 65535 -i eth0 -f on -r httpri -s \"rr test\"";
	char expected_str_2[] = "-A 192.168.1.1:80 -i eth0 -M 255.255.255.0 -m aa:bb:cc:dd:ee:ff -p 65535 -f on -r httpri -s \"rr test\" \n";
	ret = str2opt(opt_str_2, &opt, INNER_OPT_PARSE);
	ret = opt2str(&opt, str, INNER_OPT_PARSE);
	cut_assert_equal_int(0, ret);
	cut_assert_equal_string(expected_str_2, str);
}

void test_str2opt_inner_parse_a()
{
	struct options opt
		= { 0, 0, "", "", "", "", 0, "", "", "", 0, 0, 0, 0, 0 };
	int ret;

	char opt_str_1[] = "-a 192.168.1.1:80 dts01/1-3";
	ret = str2opt(opt_str_1, &opt, INNER_OPT_PARSE);
	struct options expected_1
		= { ADD_NODE, 0, "192.168.1.1:80", "", "", "", 0, "", "", "dts01/1-3", 0, 0, 0, 0, 0 };
	cut_assert_equal_int(0, ret);
	assert_opt(&expected_1, &opt);
}

void test_opt2str_inner_parse_a()
{
	struct options opt
		= { 0, 0, "", "", "", "", 0, "", "", "", 0, 0, 0, 0, 0 };
	int ret;
	char str[65535];

	char opt_str_1[] = "-a 192.168.1.1:80 dts01/1-3";
	str2opt(opt_str_1, &opt, INNER_OPT_PARSE);
	char expected_str_1[] = "-a \"192.168.1.1:80\" \"dts01/1-3\"\n";
	ret = str2opt(opt_str_1, &opt, INNER_OPT_PARSE);
	ret = opt2str(&opt, str, INNER_OPT_PARSE);
	cut_assert_equal_int(0, ret);
	cut_assert_equal_string(expected_str_1, str);
}

void test_str2opt_inner_parse_d()
{
	struct options opt
		= { 0, 0, "", "", "", "", 0, "", "", "", 0, 0, 0, 0, 0 };
	int ret;

	char opt_str_1[] = "-d 192.168.1.1:80 dts01/1-3";
	ret = str2opt(opt_str_1, &opt, INNER_OPT_PARSE);
	struct options expected_1
		= { DEL_NODE, 0, "192.168.1.1:80", "", "", "", 0, "", "", "dts01/1-3", 0, 0, 0, 0, 0 };
	cut_assert_equal_int(0, ret);
	assert_opt(&expected_1, &opt);
}

void test_opt2str_inner_parse_d()
{
	struct options opt
		= { 0, 0, "", "", "", "", 0, "", "", "", 0, 0, 0, 0, 0 };
	int ret;
	char str[65535];

	char opt_str_1[] = "-d 192.168.1.1:80 dts01/1-3";
	str2opt(opt_str_1, &opt, INNER_OPT_PARSE);
	char expected_str_1[] = "-d \"192.168.1.1:80\" \"dts01/1-3\"\n";
	ret = str2opt(opt_str_1, &opt, INNER_OPT_PARSE);
	ret = opt2str(&opt, str, INNER_OPT_PARSE);
	cut_assert_equal_int(0, ret);
	cut_assert_equal_string(expected_str_1, str);
}

void test_str2opt_inner_parse_L()
{
	struct options opt
		= { 0, 0, "", "", "", "", 0, "", "", "", 0, 0, 0, 0, 0 };
	int ret;

	/* only -L */
	char opt_str_1[] = "-L";
	ret = str2opt(opt_str_1, &opt, INNER_OPT_PARSE);
	struct options expected_1
		= { LIST, 0, "", "", "", "", 0, "", "", "", 0, 0, 0, 0, 0 };
	cut_assert_equal_int(0, ret);
	assert_opt(&expected_1, &opt);
	clear_opt(&opt);

	/* with -n */
	char opt_str_2[] = "-L -n";
	ret = str2opt(opt_str_2, &opt, INNER_OPT_PARSE);
	struct options expected_2
		= { LIST, 0, "", "", "", "", 0, "", "", "", 0, 0, 1, 0, 0 };
	cut_assert_equal_int(0, ret);
	assert_opt(&expected_2, &opt);
	clear_opt(&opt);

	/* with -c */
	char opt_str_3[] = "-L -c";
	ret = str2opt(opt_str_3, &opt, INNER_OPT_PARSE);
	struct options expected_3
		= { LIST, 0, "", "", "", "", 0, "", "", "", 0, 1, 0, 0, 0 };
	cut_assert_equal_int(0, ret);
	assert_opt(&expected_3, &opt);
	clear_opt(&opt);

	/* -n and -c */
	char opt_str_4[] = "-L -n -c";
	ret = str2opt(opt_str_4, &opt, INNER_OPT_PARSE);
	struct options expected_4
		= { LIST, 0, "", "", "", "", 0, "", "", "", 0, 1, 1, 0, 0 };
	cut_assert_equal_int(0, ret);
	assert_opt(&expected_4, &opt);
}

void test_opt2str_inner_parse_L()
{
	struct options opt
		= { 0, 0, "", "", "", "", 0, "", "", "", 0, 0, 0, 0, 0 };
	int ret;
	char str[65535];

	char opt_str_1[] = "-L";
	str2opt(opt_str_1, &opt, INNER_OPT_PARSE);
	char expected_str_1[] = "-L  \n";
	ret = str2opt(opt_str_1, &opt, INNER_OPT_PARSE);
	ret = opt2str(&opt, str, INNER_OPT_PARSE);
	cut_assert_equal_int(0, ret);
	cut_assert_equal_string(expected_str_1, str);
	clear_opt(&opt);

	char opt_str_2[] = "-L -n";
	str2opt(opt_str_2, &opt, INNER_OPT_PARSE);
	char expected_str_2[] = "-L  -n\n";
	ret = str2opt(opt_str_2, &opt, INNER_OPT_PARSE);
	ret = opt2str(&opt, str, INNER_OPT_PARSE);
	cut_assert_equal_int(0, ret);
	cut_assert_equal_string(expected_str_2, str);
	clear_opt(&opt);

	char opt_str_3[] = "-L -c -n";
	str2opt(opt_str_3, &opt, INNER_OPT_PARSE);
	char expected_str_3[] = "-L -c -n\n";
	ret = str2opt(opt_str_3, &opt, INNER_OPT_PARSE);
	ret = opt2str(&opt, str, INNER_OPT_PARSE);
	cut_assert_equal_int(0, ret);
	cut_assert_equal_string(expected_str_3, str);
	clear_opt(&opt);
}

void test_str2opt_inner_parse_G()
{
	struct options opt
		= { 0, 0, "", "", "", "", 0, "", "", "", 0, 0, 0, 0, 0 };
	int ret;

	char opt_str_1[] = "-G";
	ret = str2opt(opt_str_1, &opt, INNER_OPT_PARSE);
	struct options expected_1
		= { GET_PROXY_PORT, 1, "", "", "", "", 0, "", "", "", 0, 0, 0, 0, 0 };
	cut_assert_equal_int(0, ret);
	assert_opt(&expected_1, &opt);
}

void test_opt2str_inner_parse_G()
{
	struct options opt
		= { 0, 0, "", "", "", "", 0, "", "", "", 0, 0, 0, 0, 0 };
	int ret;
	char str[65535];

	char opt_str_1[] = "-G";
	str2opt(opt_str_1, &opt, INNER_OPT_PARSE);
	char expected_str_1[] = "-G\n";
	ret = str2opt(opt_str_1, &opt, INNER_OPT_PARSE);
	ret = opt2str(&opt, str, INNER_OPT_PARSE);
	cut_assert_equal_int(0, ret);
	cut_assert_equal_string(expected_str_1, str);
}

void test_str2opt_inner_parse_B()
{
	struct options opt
		= { 0, 0, "", "", "", "", 0, "", "", "", 0, 0, 0, 0, 0 };
	int ret;

	char opt_str_1[] = "-B -p 65535";
	ret = str2opt(opt_str_1, &opt, INNER_OPT_PARSE);
	struct options expected_1
		= { BIND_SOCK, 1, "", "", "", "", 0, "", "", "", 65535, 0, 0, 0, 0 };
	cut_assert_equal_int(0, ret);
	assert_opt(&expected_1, &opt);
}

void test_opt2str_inner_parse_B()
{
	struct options opt
		= { 0, 0, "", "", "", "", 0, "", "", "", 0, 0, 0, 0, 0 };
	int ret;
	char str[65535];

	char opt_str_1[] = "-B -p 65535";
	str2opt(opt_str_1, &opt, INNER_OPT_PARSE);
	char expected_str_1[] = "-B -p 65535\n";
	ret = str2opt(opt_str_1, &opt, INNER_OPT_PARSE);
	ret = opt2str(&opt, str, INNER_OPT_PARSE);
	cut_assert_equal_int(0, ret);
	cut_assert_equal_string(expected_str_1, str);
}

void test_str2opt_inner_parse_U()
{
	struct options opt
		= { 0, 0, "", "", "", "", 0, "", "", "", 0, 0, 0, 0, 0 };
	int ret;

	char opt_str_1[] = "-U -p 65535";
	ret = str2opt(opt_str_1, &opt, INNER_OPT_PARSE);
	struct options expected_1
		= { UNBIND_SOCK, 1, "", "", "", "", 0, "", "", "", 65535, 0, 0, 0, 0 };
	cut_assert_equal_int(0, ret);
	assert_opt(&expected_1, &opt);
}

void test_opt2str_inner_parse_U()
{
	struct options opt
		= { 0, 0, "", "", "", "", 0, "", "", "", 0, 0, 0, 0, 0 };
	int ret;
	char str[65535];

	char opt_str_1[] = "-U -p 65535";
	str2opt(opt_str_1, &opt, INNER_OPT_PARSE);
	char expected_str_1[] = "-U -p 65535\n";
	ret = str2opt(opt_str_1, &opt, INNER_OPT_PARSE);
	ret = opt2str(&opt, str, INNER_OPT_PARSE);
	cut_assert_equal_int(0, ret);
	cut_assert_equal_string(expected_str_1, str);
}

void test_str2opt_inner_parse_H()
{
	struct options opt
		= { 0, 0, "", "", "", "", 0, "", "", "", 0, 0, 0, 0, 0 };
	int ret;

	char opt_str_1[] = "-H -i eth0";
	ret = str2opt(opt_str_1, &opt, INNER_OPT_PARSE);
	struct options expected_1
		= { SET_MACADDR, 1, "", "eth0", "", "", 0, "", "", "", 0, 0, 0, 0, 0 };
	cut_assert_equal_int(0, ret);
	assert_opt(&expected_1, &opt);
}

void test_opt2str_inner_parse_H()
{
	struct options opt
		= { 0, 0, "", "", "", "", 0, "", "", "", 0, 0, 0, 0, 0 };
	int ret;
	char str[65535];

	char opt_str_1[] = "-H -i eth0";
	str2opt(opt_str_1, &opt, INNER_OPT_PARSE);
	char expected_str_1[] = "-H -i eth0\n";
	ret = str2opt(opt_str_1, &opt, INNER_OPT_PARSE);
	ret = opt2str(&opt, str, INNER_OPT_PARSE);
	cut_assert_equal_int(0, ret);
	cut_assert_equal_string(expected_str_1, str);
}

