#include <cutter.h>
#include <stdint.h>
#include <string.h>
#include <pthread.h>
#include <unistd.h>
#include <stdio.h>
#include <alloca.h>

struct hbc_node
{
	const char *name;
	uint32_t addr;
	int ready;
};

struct hbc_node_container
{
	struct hbc_node_container * next;
	struct hbc_node * hbc_node;
};

extern struct hbc_node * find_hbc_node(const char * name);
extern int add_hbc_node(const struct hbc_node new_node);
extern int update_hbc_node(const struct hbc_node new_node);

void test_find_node (void);

void
test_new_stack (void)
{
	cut_assert(find_hbc_node("hoge") == NULL);
}

extern int parse_idset(char * str, uint64_t * IDset);

#define idof(a) (1ull << a)
#define idset_instance1 "1,2,3,4"
#define idset_instance2 "1-4"
#define idset_instance3 "1,2,3,10-14"
#define idset_instance4 "4,5,6,7"
#define idset_instance5 "60-63"

void
parse_idset1() {
	uint64_t IDset = idof(1) | idof(2) | idof(3) | idof(4), ret;
	char target[] = idset_instance1;
	parse_idset(target, &ret);
	cut_assert(ret == IDset);
}

void
test_parse_idset2() {
	uint64_t IDset = idof(1) | idof(2) | idof(3) | idof(4), ret;
	char target[] = idset_instance2;
	parse_idset(target, &ret);
	cut_assert(ret == IDset);
}

void
test_parse_idset3() {
	uint64_t IDset = idof(1) | idof(2) | idof(3) | idof(10) | idof(11) | idof(12) | idof(13) | idof(14), ret;
	char target[] = idset_instance3;
	parse_idset(target, &ret);
	cut_assert(ret == IDset);
}

void
test_parse_idset4() {
	uint64_t IDset = idof(4) | idof(5) | idof(6) | idof(7), ret;
	char target[] = idset_instance4;
	parse_idset(target, &ret);
	cut_assert(ret == IDset);
}

void
test_parse_idset5() {
	uint64_t IDset = idof(60) | idof(61) | idof(62) | idof(63), ret;
	char target[] = idset_instance5;
	parse_idset(target, &ret);
	cut_assert(ret == IDset);
}


extern int is_emptyQueue();
extern void enqueue(const char * result);
extern char * dequeue();
extern void clear_queue();
extern int count_queue();
extern void print_queue();


void test_queue_first() {
	cut_assert_equal_int(1, is_emptyQueue());
}

void test_enqueue1() {
	enqueue("first");
	char * de = dequeue();
	cut_assert_equal_string("first", de);
}

void test_enqueue2() {
	char *de;
	cut_assert_equal_int(0, count_queue());
	enqueue("first");
	cut_assert_equal_int(1, count_queue());
	enqueue("second");
	cut_assert_equal_int(2, count_queue());
	enqueue("third");
	cut_assert_equal_int(3, count_queue());
	enqueue("forth");
	cut_assert_equal_int(4, count_queue());

	de = dequeue();
	cut_assert_equal_string("first", de);

	de = dequeue();
	cut_assert_equal_string("second", de);

	de = dequeue();
	cut_assert_equal_string("third", de);

	clear_queue();

	de = dequeue();
	cut_assert(NULL == de);
}

extern int timed_is_emptyQueue();
extern int timed_enqueue(const char * result);
extern char * timed_dequeue();
extern void timed_clear_queue();
extern int timed_count_queue();
extern void timed_print_queue();

void test_timed_enqueue() {
	char *de;
	timed_enqueue("first");
	timed_enqueue("second");
	timed_enqueue("third");
	timed_enqueue("forth");

	de = timed_dequeue();
	cut_assert(NULL == de);
}

struct queue_timer_arguments {
	struct timespec timeout;
	struct timespec stride;
	int count;
};

void test_timed_enqueue2() {
	char *de;
	pthread_t th;

	struct queue_timer_arguments args = {
		.timeout = {.tv_sec = 2, .tv_nsec = 0, },
		.stride = {.tv_sec = 0, .tv_nsec = 10000000, },
		.count = 3,
	};

	start_receive_result(&th, &args);
	timed_enqueue("first");
	timed_enqueue("second");
	timed_enqueue("third");
	timed_enqueue("forth");
	wait_all_request(th);

	de = timed_dequeue();
	cut_assert_equal_string("first", de);

	de = timed_dequeue();
	cut_assert_equal_string("second", de);

	de = timed_dequeue();
	cut_assert_equal_string("third", de);

	de = timed_dequeue();
	cut_assert(NULL == de);
}

void test_timed_enqueue3() {
	char *de;
	pthread_t th;

	struct queue_timer_arguments args = {
		.timeout = {.tv_sec = 1, .tv_nsec = 0, },
		.stride = {.tv_sec = 0, .tv_nsec = 10000000, },
		.count = 3,
//		.count = 4,
	};

	start_receive_result(&th, &args);
	timed_enqueue("first");
	timed_enqueue("second");
	timed_enqueue("third");
	//sleep(2);
	//timed_enqueue("forth");
	wait_all_request(th);

	de = timed_dequeue();
	cut_assert_equal_string("first", de);

	de = timed_dequeue();
	cut_assert_equal_string("second", de);

	de = timed_dequeue();
	cut_assert_equal_string("third", de);

	de = timed_dequeue();
	cut_assert(NULL == de);
}

void test_hbcNodeInterface()
{
	struct hbc_node new_node;
	struct hbc_node * node;

	new_node.name = "node01"; // avoid warning
	new_node.addr= 144;
	new_node.ready = 1;
	add_hbc_node(new_node);
	new_node.name = "node02"; // avoid warning
	new_node.addr= 144;
	new_node.ready = 1;
	add_hbc_node(new_node);
	node = find_hbc_node("node01");
	cut_assert_equal_string("node01", node->name);
	node = find_hbc_node("node02");
	cut_assert_equal_string("node02", node->name);

	node = find_hbc_node("node03");
	cut_assert(node == NULL);

}

extern int hello_world(int * status, char ** res, char ** err);

void test_echo_hello_world_command()
{
	int status;
	char res[4092];
	char err[4092];

	res [0] = '\0';

	hello_world(&status, (char*)res, (char*)err);

	cut_assert_equal_string("Hello World\n", res);
}

#include <dtsd_common.h>
#include <dtsd_optparser.h>
#include <dtsd_server_conf.h>

void test_resolve_host()
{
	int ret;
	uint32_t addr;
	uint32_t expected = 0x0100a8c0;
	const char address_string[] = "192.168.0.1";

	ret = resolve_host(address_string, &addr);
	// fprintf(stderr, "addr = %x\n", addr);

	cut_assert_equal_int(0, ret);
	cut_assert_equal_int(expected, addr);
}

void test_parse_hostsrv()
{
	int ret;
	uint32_t addr;
	uint32_t expected_addr = 0x1e00a8c0;
	int port;
	char hstsrv[] = "192.168.0.30:80";

	ret = parse_hostsrv(hstsrv, &addr, &port);
	cut_assert_equal_int(0, ret);
	cut_assert_equal_int(expected_addr, addr);
	cut_assert_equal_int(80, port);
}

void test_str2opt()
{
	struct options opt;
	const char str[] = "-A 192.168.0.30:9999 -i eth0 -M 255.255.255.0 -m a4:ba:db:03:7e:7e -p 60000 -f on -r httpri -s \"-d rr\" ";
	cut_assert(str2opt(str, &opt, INNER_OPT_PARSE) == 0);
}

void test_servConf_NodeInfo()
{
	char message[MAX_MSG_SIZE];

	int len;
	struct options opt;
	struct server_conf conf = {
		.next = NULL,
		.server_address = 0x1e00a8c0,
		.netmask = 0x00ffffff,
		.server_port = 80,
		.iface = "eth0",
		.hwaddr = "a4:ba:db:03:7e:7e",
		.fakemac = 1,
		.riname = "httpri",
		.ri = NULL,
		.ri_intf = NULL,
		.ri_opt = "-d rr",
		.ni = NULL,
		.proxy_port = 50000,
		.proxy_sock = -1,
	};
	struct server_conf * p1, * p2, * r1 = NULL;
	struct dtsd_node_info info = {
		.next = NULL,
		.srv_conf = &conf,
		.hb_node_name = "dts01",
		.hbc_node = NULL,
		.idset = 0x000000000000000f,
	};
	struct dtsd_node_info *i1, *i2, *i3;

	p1 = (struct server_conf*) malloc ((sizeof(struct server_conf)));
	p2 = (struct server_conf*) malloc ((sizeof(struct server_conf)));

	ServerConf2opt(&conf, (struct options *)(&opt));
	cut_assert_equal_int(ADD_CONF, opt.cmd);
	cut_assert_equal_string("192.168.0.30:80", opt.host_srv);
	cut_assert_equal_string("255.255.255.0", opt.netmask);
	cut_assert_equal_string("eth0", opt.iface);
	cut_assert_equal_string("httpri", opt.ri);
	cut_assert_equal_string("-d rr", opt.ri_opt);
	cut_assert_equal_string("a4:ba:db:03:7e:7e", opt.macaddr);
	cut_assert_equal_int(50000, opt.proxy_port);

	opt2ServerConf(&opt, p1);
	cut_assert_equal_int(0x1e00a8c0, p1->server_address);
	cut_assert_equal_int(80, p1->server_port);
	cut_assert_equal_int(0x00ffffff, p1->netmask);
	cut_assert_equal_string("eth0", p1->iface);
	cut_assert_equal_string("httpri", p1->riname);
	cut_assert_equal_string("-d rr", p1->ri_opt);
	cut_assert_equal_string("a4:ba:db:03:7e:7e", p1->hwaddr);
	cut_assert_equal_int(50000, p1->proxy_port);

	free_server_conf(p1);
	cut_assert_equal_string("192.168.0.30:80", opt.host_srv);
	cut_assert_equal_string("255.255.255.0", opt.netmask);
	cut_assert_equal_string("eth0", opt.iface);
	cut_assert_equal_string("httpri", opt.ri);
	cut_assert_equal_string("-d rr", opt.ri_opt);
	cut_assert_equal_string("a4:ba:db:03:7e:7e", opt.macaddr);
	cut_assert_equal_int(50000, opt.proxy_port);

	p1 = NULL;
	p1 = (struct server_conf*) malloc ((sizeof(struct server_conf)));
	opt2ServerConf(&opt, p1);
	strcpy(opt.host_srv, "192.168.0.40:80");
	opt2ServerConf(&opt, p2);
	p1->server_port = 9999;

	append_ServerConfList(&conf, p1);
	append_ServerConfList(&conf, p2);

	NodeInfo2opt(&info, &opt);
	cut_assert_equal_int(ADD_NODE, opt.cmd);
	cut_assert_equal_string("192.168.0.30:80", opt.host_srv);
	cut_assert_equal_string("dts01/0,1,2,3", opt.nodes);
	// TODO? cut_assert_equal_string("dts01/0-3", opt.nodes);

	i1 = (struct dtsd_node_info *) malloc ((sizeof(struct dtsd_node_info)));
	i1->hb_node_name = NULL;
	opt2NodeInfo(&opt, i1);
	cut_assert_equal_string("dts01", i1->hb_node_name);
	cut_assert_equal_int(0x000000000000000f, i1->idset);

	strcpy(opt.nodes, "dts02/4,5,6,7");
	i2 = (struct dtsd_node_info *) malloc ((sizeof(struct dtsd_node_info)));
	i2->hb_node_name = NULL;
	opt2NodeInfo(&opt, i2);
	cut_assert_equal_string("dts02", i2->hb_node_name);
	cut_assert_equal_int(0x00000000000000f0, i2->idset);

	cut_assert(i1->srv_conf == NULL);
	cut_assert(i2->srv_conf == NULL);

	add_NodeInfo(p1, i1);
	add_NodeInfo(p1, i2);

	cut_assert(conf.next == p1);
	cut_assert(conf.next->next == p2);
	cut_assert(p2->next == NULL);
	cut_assert(p1->ni == i1);
	cut_assert(p1->ni->next == i2);
	cut_assert(p1->ni->srv_conf == p1);
	cut_assert(p1->ni->next->srv_conf == p1);
	cut_assert_equal_int(0x00000000000000f0, p1->ni->next->idset);

	cut_assert(ServerConf2str(p1, message, sizeof(message)) == 0);
	cut_assert(str2ServerConf(message, &r1) == 0);

	cut_assert(r1 != NULL);
	cut_assert_equal_int(0x1e00a8c0, r1->server_address);
	cut_assert_equal_int(9999, r1->server_port);
	cut_assert_equal_int(0x00ffffff, r1->netmask);
	cut_assert(r1->ni != NULL);
	cut_assert_equal_string(r1->ni->hb_node_name, "dts01");
	cut_assert_equal_string(r1->ni->next->hb_node_name, "dts02");

	len = strlen(message);
	cut_assert(ServerConf2str(p1, message, len) == 0);
	cut_assert(ServerConf2str(p1, message, len - 1) != 0);

	free_server_conf_list(p1);
	
}

extern int add_socket(int port, int socket);
extern int fetch_socket(const int port, int * socket);
extern int bind_socket(const int port);

void test_add_socket()
{
	int s = -1;
	cut_assert_equal_int(0, add_socket(80, 1));
	cut_assert_equal_int(0, add_socket(99, 21));
	cut_assert_equal_int(0, fetch_socket(80, &s));
	cut_assert_equal_int(1, s);
	cut_assert_equal_int(0, fetch_socket(99, &s));
	cut_assert_equal_int(21, s);
	cut_assert_equal_int(-1, fetch_socket(99, &s));
}

void test_bind_socket()
{
	int s = -1;
	cut_assert_equal_int(0, bind_socket(60080));
	cut_assert_equal_int(0, bind_socket(60099));
	cut_assert_equal_int(0, fetch_socket(60080, &s));
	cut_assert_not_equal_int(-1, s);
	close(s);
	s = -1;
	cut_assert_equal_int(0, fetch_socket(60099, &s));
	cut_assert_not_equal_int(-1, s);
	close(s);
	cut_assert_equal_int(-1, fetch_socket(60099, &s));
}
