/*
  Simple binary-tree implementation

  This is a simplified version of rbtree.c in Linux kernel,
  written by Andrea Arcangeli.
 */

#include "common.h"

void bt_link_node(struct bt_node *new, struct bt_node *parent,
				  struct bt_node **link)
{
	new->left = new->right = NULL;
	new->parent = parent;

	*link = new;
}

struct bt_node *bt_first(const struct bt_root *root)
{
	struct bt_node *node = root->node;

	if (node == NULL)
		return NULL;

	while (node->left)
		node = node->left;

	return node;
}

struct bt_node *bt_last(const struct bt_root *root)
{
	struct bt_node *node = root->node;

	if (node == NULL)
		return NULL;

	while (node->right)
		node = node->right;

	return node;
}

struct bt_node *bt_prev(const struct bt_node *node)
{
	if (node->left) {
		while (node->left)
			node = node->left;
		return node;
	}

	if (node->parent == NULL)
		return NULL;

	while (node->parent && node->parent->left == node)
		node = node->parent;

	return node;
}

struct bt_node *bt_next(const struct bt_node *node)
{
	if (node->right) {
		while (node->right)
			node = node->right;

		return node;
	}

	if (node->parent == NULL)
		return NULL;

	while (node->parent && node->parent->right == node)
		node = node->parent;

	return node;
}
