#ifndef HTTPRI_COMMON_H
#define HTTPRI_COMMON_H

#include <stdio.h>
#include <stdint.h>
#include <stddef.h>

#include "../dtsd/src/dtsd_ri_plugins.h"

int verbous;
#define dprintf(level, fmt...)									\
	do { if (verbous >= level) fprintf(stderr, fmt); } while(0)
#define dprints(level, str, len)										\
	do { if (verbous >= level) fprints(stderr, str, len); } while (0)

#define HTTPRI_DEFAULT_PORT 8080
#define HTTPRI_DEFAULT_BUFSIZE 2048
#define DEFAULT_BACKLOG 500
#define HTTPRI_MAX_TARGETS 64

struct http_request {
	const char *method;
	int method_len;
	const char *uri;
	int uri_len;
	const char *version;
	int version_len;
};

struct http_header {
	const char *name;
	int name_len;
	const char *value;
	int value_len;
};

#define HTTPRI_IPADDR_INVALID (0xffffffff)

struct httpri_dispatcher {
	const char *name;
	int (*init)(struct ri * ri, uint32_t default_target);
	uint32_t (*dispatch)(struct ri * ri, const struct http_request *req);
	int (*add_node)(struct ri * ri, uint32_t addr);
	void (*remove_node)(struct ri * ri, uint32_t addr);
	void * node_info;
	void (*free_node_info)(struct ri * ri);
};

struct httpri_dispatcher_mgr {
	const char *name;
	struct httpri_dispatcher *(*create)(const char *name);
};

struct httpri_uniq_param {
	struct httpri_dispatcher * dispatcher;
	pthread_mutex_t node_list_lock;
	struct server_param * server_param;
	int n_threads;
	struct thread_param * thread_params;
};
	
extern struct httpri_dispatcher_mgr rr_dispatcher_mgr;
//extern struct httpri_dispatcher_mgr specweb_dispatcher_mgr;

/* Taken from linux/kernel.h */
/**
 * container_of - cast a member of a structure out to the containing structure
 * @ptr:	the pointer to the member.
 * @type:	the type of the container struct this is embedded in.
 * @member:	the name of the member within the struct.
 *
 */
#define container_of(ptr, type, member) ({			\
	const typeof( ((type *)0)->member ) *__mptr = (ptr);	\
	(type *)( (char *)__mptr - offsetof(type,member) );})

struct bt_node {
	struct bt_node *parent;
	struct bt_node *left;
	struct bt_node *right;
};

struct bt_root {
	struct bt_node *node;
};

void bt_link_node(struct bt_node *new, struct bt_node *parent,
				  struct bt_node **link);
struct bt_node *bt_first(const struct bt_root *root);
struct bt_node *bt_last(const struct bt_root *root);
struct bt_node *bt_prev(const struct bt_node *node);
struct bt_node *bt_next(const struct bt_node *node);

#endif /* #ifndef HTTPRI_COMMON_H */
