#include <stdlib.h>
#include <netdb.h>
#include <stdint.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>

#include "common.h"

#define MAX_TARGETS 128

static void print_usage(void)
{
	fprintf(stderr,
			"Usage: specweb_test [options] \n"
			"\n"
			"Options:\n"
			"  -d <dispatcher>  specifies a dispatcher\n"
			"  -t <target>      specifies a target address\n"
			"  -r <target>      specifies a target address to be removed later\n"
			"  -x <N>           remove node after processing N requests\n");
}

static void remove_newline(char *buff, size_t len)
{
	if (len > 0 && buff[len-1] == '\n')
		buff[len-1] = '\0';
}

static struct httpri_dispatcher *select_dispatcher(const char *name)
{
	if (strcmp(name, "rr") == 0)
		return &rr_dispatcher;
	if (strcmp(name, "specweb") == 0)
		return &specweb_dispatcher;

	return NULL;
}

int main(int argc, char *argv[])
{
	char buff[BUFSIZ];
	uint32_t target_addrs[MAX_TARGETS];
	int targets = 0, i;
	uint32_t remove_addr = 0;
	struct httpri_dispatcher *disp = NULL;
	int processed = 0;
	int remove_after = 0;

	for (;;) {
		int r;
		struct hostent *ent;

		r = getopt(argc, argv, "d:r:t:x:");
		if (r == -1)
			break;

		switch (r) {
			uint32_t addr;

		case 'd':
			disp = select_dispatcher(optarg);
			if (disp == NULL) {
				fprintf(stderr, "No such dispatcher: %s\n", optarg);
				return 1;
			}
			break;

		case 'r':
		case 't':
			if (targets >= MAX_TARGETS) {
				fprintf(stderr, "Too many target hosts! Sorry!\n");
				return 1;
			}

			ent = gethostbyname(optarg);
			if (ent == NULL) {
				fprintf(stderr, "Unknown hostname: %s\n", optarg);
				return 1;
			}
			memcpy(&addr, ent->h_addr, sizeof(uint32_t));

			if (r == 'r')
				remove_addr = addr;
			else
				target_addrs[targets++] = addr;
			break;

		case 'x':
			remove_after = atoi(optarg);
			break;

		case '?':
			print_usage();
			return 1;
		}
	}

	if (disp == NULL) {
		fprintf(stderr, "No dispatcher specified\n");
		return 1;
	}

	if (disp->init(0)) { /* TOOD! */
		fprintf(stderr, "disp->init() failed\n");
		return 1;
	}

	for (i = 0; i < targets; i++) {
		if (disp->add_node(target_addrs[i])) {
				fprintf(stderr, "disp->add_node failed\n");
				return 1;
		}
	}

	while (fgets(buff, BUFSIZ, stdin)) {
		struct http_request req;
		size_t len = strlen(buff);
		uint32_t addr;
		struct in_addr in;

		if (remove_addr != 0 && processed == remove_after) {
			fprintf(stderr, "removing node\n");
			disp->remove_node(remove_addr);
		}


		remove_newline(buff, len);
		req.uri = buff;
		req.uri_len = len;

		addr = disp->dispatch(&req);
		in.s_addr = addr;
		printf("%s\t%s\n", buff, inet_ntoa(in));

		processed++;
	}

	return 0;
}
