#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>

#include "ftcs.h"
#include "Python.h"

static PyObject*
request_handover(PyObject* self, PyObject *args)
{
    char *dest_host, *buf;
    int buflen, sd, rv, port, offset;
    struct sockaddr_in addr;
    struct hostent *ent;
    char errmsg[BUFSIZ];

    rv = PyArg_ParseTuple(args, "isis#i", &sd, &dest_host, &port, &buf, &buflen, &offset);
    if (!rv) {
        return NULL;
    }
    ent = gethostbyname(dest_host);
    if (ent == NULL) {
        snprintf(errmsg, sizeof(errmsg),
            "gethostbyname() failed: %s", strerror(errno));
        PyErr_SetString(PyExc_RuntimeError, errmsg);
        return NULL;
    }
    addr.sin_family = AF_INET;
    addr.sin_port   = htons(port);
    memcpy(&addr.sin_addr.s_addr, ent->h_addr, ent->h_length);

    Py_BEGIN_ALLOW_THREADS
    rv = ftcs_request_handover(sd, (struct sockaddr*)&addr, sizeof(addr),
        buf, buflen, offset);
    Py_END_ALLOW_THREADS
    if (rv < 0) {
        snprintf(errmsg, sizeof(errmsg),
            "ftcs_request_handover() failed: %s, "
            "sd:%d host:%s port:%d buflen:%d offset:%d",
            strerror(errno),
            sd, dest_host, port, buflen, offset);
        PyErr_SetString(PyExc_RuntimeError, errmsg);
        return NULL;
    }
    return Py_BuildValue("i", rv);
}

static PyMethodDef methods[] = {
    {"request_handover", request_handover, METH_VARARGS, "call ftcs_request_handover()"},
    {NULL, NULL, 0, NULL}
};

PyMODINIT_FUNC
init_ftcs(void)
{
    (void) Py_InitModule("_ftcs", methods);
}
